import contextlib
import json
import logging
import os
import subprocess
import uuid

import tqdm
import yaml
import znflow

from zntrack import utils
from zntrack.group import Group
from zntrack.state import PLUGIN_LIST
from zntrack.utils.finalize import make_commit
from zntrack.utils.import_handler import import_handler
from zntrack.utils.misc import load_env_vars

from . import config
from .deployment import ZnTrackDeployment

log = logging.getLogger(__name__)


class _FinalNodeNameString(str):
    """A string that represents the final name of a node.

    Used to differentiate between a custom name and a computed name.
    """


class Project(znflow.DiGraph):
    def __init__(
        self,
        *args,
        disable=False,
        immutable_nodes=True,
        deployment=None,
        tags: dict[str, str] | None = None,
        **kwargs,
    ):
        if deployment is None:
            deployment = ZnTrackDeployment()
        self.tags = tags or {}
        load_env_vars()
        super().__init__(
            *args,
            disable=disable,
            immutable_nodes=immutable_nodes,
            deployment=deployment,
            **kwargs,
        )

    def compute_all_node_names(self) -> dict[uuid.UUID, str]:
        """Compute the Node name based on existing nodes on the graph."""
        all_nodes = [self.nodes[uuid]["value"] for uuid in self.nodes]
        node_names = {}
        for node in all_nodes:
            custom_name = node.__dict__.get("name")
            node_name = custom_name or node.__class__.__name__
            if custom_name is not None and isinstance(custom_name, _FinalNodeNameString):
                node_names[node.uuid] = custom_name
                continue

            if node.state.group is None:
                if self.active_group is not None:
                    node_name = f"{'_'.join(self.active_group.names)}_{node_name}"
                else:
                    node_name = f"{node_name}"
            else:
                node_name = f"{'_'.join(node.state.group.names)}_{node_name}"

            if node_name in node_names.values():
                if custom_name:
                    raise ValueError(
                        f"A node with the name '{node_name}' already exists."
                    )
                i = 0
                while True:
                    i += 1
                    if f"{node_name}_{i}" not in node_names.values():
                        node_name = f"{node_name}_{i}"
                        break
            node_names[node.uuid] = _FinalNodeNameString(node_name)

        return node_names

    def add_node(self, node_for_adding, **attr):
        from zntrack import Node

        if not isinstance(node_for_adding, Node):
            raise ValueError(
                f"Node must be an instance of 'zntrack.Node', not {type(node_for_adding)}"
            )

        return super().add_node(node_for_adding, **attr)

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            for group in self.groups.values():
                for node_uuid in group.uuids:
                    # we need to access the `state` attribute to initialize
                    # the property, so the `log.debug` is necessary!
                    log.debug(self.nodes[node_uuid]["value"].state)
                    self.nodes[node_uuid]["value"].__dict__["state"]["group"] = (
                        Group.from_znflow_group(group)
                    )

            all_node_names = self.compute_all_node_names()
            for node_uuid in self.nodes:
                self.nodes[node_uuid]["value"].__dict__["name"] = all_node_names[
                    node_uuid
                ]
        finally:
            super().__exit__(exc_type, exc_val, exc_tb)

    def build(self) -> None:
        log.info(f"Saving {config.PARAMS_FILE_PATH}")
        params_dict = {}
        dvc_dict = {"stages": {}, "plots": []}
        zntrack_dict = {}
        for node_uuid in tqdm.tqdm(self):
            node = self.nodes[node_uuid]["value"]
            if node._external_:
                continue
            for plugin in node.state.plugins.values():
                # TODO: combine all params into one dict
                if (
                    value := plugin.convert_to_params_yaml()
                ) is not config.PLUGIN_EMPTY_RETRUN_VALUE:
                    params_dict[node.name] = value
                if (
                    value := plugin.convert_to_dvc_yaml()
                ) is not config.PLUGIN_EMPTY_RETRUN_VALUE:
                    dvc_dict["stages"][node.name] = value["stages"]
                    # TODO: this won't work if multiple
                    # plugins want to modify the dvc.yaml
                    if len(value["plots"]) > 0:
                        dvc_dict["plots"].extend(value["plots"])
                if (
                    value := plugin.convert_to_zntrack_json(graph=self)
                ) is not config.PLUGIN_EMPTY_RETRUN_VALUE:
                    zntrack_dict[node.name] = value

        if len(dvc_dict["plots"]) == 0:
            del dvc_dict["plots"]

        config.PARAMS_FILE_PATH.write_text(yaml.safe_dump(params_dict))
        config.DVC_FILE_PATH.write_text(yaml.safe_dump(dvc_dict))
        config.ZNTRACK_FILE_PATH.write_text(json.dumps(zntrack_dict, indent=4))

        # TODO: update file or overwrite?

    def repro(self, build: bool = True, force: bool = False):
        if build:
            self.build()
        cmd = ["dvc", "repro"]
        if force:
            cmd.append("--force")
        subprocess.check_call(cmd)

    def finalize(
        self,
        msg: str | None = None,
        commit: bool = True,
        skip_cached: bool = True,
        update_run_names: bool = True,
        **kwargs,
    ):
        """Finalize the project by making a commit and loading environment variables.

        This method performs the following actions:
        1. Makes a commit with the provided message if `commit` is True.
        2. Loads environment variables.
        3. Loads and finalizes plugins specified
           in the `ZNTRACK_PLUGINS` environment variable.

        Parameters
        ----------
        msg : str, optional
            The commit message. If None, the message will be 'zntrack: auto commit'.
        commit : bool, optional
            Whether to make a commit or not. Default is True
        skip_cached : bool, optional
            Do not upload cached nodes.
        update_run_names: bool, optional
            Include part of the commit message in the run names.
        **kwargs
            Additional keyword arguments to pass to `make_commit`.

        """
        if msg is None:
            msg = "zntrack: auto commit"
        if commit:
            make_commit(msg, **kwargs)
        utils.misc.load_env_vars()
        plugins_paths = os.environ.get(
            "ZNTRACK_PLUGINS", "zntrack.plugins.dvc_plugin.DVCPlugin"
        )
        plugins: PLUGIN_LIST = [import_handler(p) for p in plugins_paths.split(",")]
        for plugin in plugins:
            plugin.finalize(skip_cached=skip_cached, update_run_names=update_run_names)

    @contextlib.contextmanager
    def group(self, *names: str):
        """Group nodes together.

        Parameters
        ----------
        names : list[str], optional
            The name of the group. If None, the group will be named 'GroupX' where X is
            the number of groups + 1. If more than one name is given, the groups will
            be nested to 'nwd = name[0]/name[1]/.../name[-1]'

        """
        # This context manager also open self.__enter__ in the `super`
        if not names:
            name = "Group1"
            while (name,) in self.groups:
                name = f"Group{int(name[5:]) + 1}"
            names = (name,)

        with super().group(*names) as znflow_grp:
            group = Group(names=znflow_grp.names)
            yield group
        group._nodes = znflow_grp.nodes



import zntrack.examples


def test_connect_node_names(proj_path):
    with zntrack.Project() as proj:
        a = zntrack.examples.ParamsToOuts(params=10, name="A")
        assert a.name == "A"
        b = zntrack.examples.ParamsToOuts(params=20, name=f"con_{a.name}")
        assert b.name == "con_A"
    proj.repro()

    assert a.name == "A"
    assert b.name == "con_A"


def test_connect_node_names_automatic_names(proj_path):
    with zntrack.Project() as proj:
        a = zntrack.examples.ParamsToOuts(params=10)
        assert a.name == "ParamsToOuts"
        b = zntrack.examples.ParamsToOuts(params=20, name=f"con_{a.name}")
        assert b.name == "con_ParamsToOuts"

    proj.repro()

    assert a.name == "ParamsToOuts"
    assert b.name == "con_ParamsToOuts"


def test_connect_nodes_in_grp(proj_path):
    proj = zntrack.Project()
    with proj.group("A", "B"):
        a = zntrack.examples.ParamsToOuts(params=10)
        assert a.name == "A_B_ParamsToOuts"
        b = zntrack.examples.ParamsToOuts(params=20, name=f"con_{a.name}")
        assert b.name == "A_B_con_A_B_ParamsToOuts"

    proj.repro()

    assert a.name == "A_B_ParamsToOuts"
    assert b.name == "A_B_con_A_B_ParamsToOuts"


# def test_path_as_name(proj_path):
#     proj = zntrack.Project()
#     with proj:
#         a = zntrack.examples.WriteDVCOuts(params=10, outs="outs.txt")
#         # This is impossible, because the connection would have to be resolved
#         # within the f_string itself.
#         b = zntrack.examples.WriteDVCOuts(params=20, name=f"con_{a.outs}")

#         assert a.name == "WriteDVCOuts"
#         assert b.name == "con_outs.txt"

#     proj.repro()



import copy
import dataclasses
import json
import pathlib
import typing as t

import znflow
import znflow.handler
import znflow.utils
import znjson

from zntrack import config, converter
from zntrack.config import (
    FIELD_TYPE,
    PLUGIN_EMPTY_RETRUN_VALUE,
    ZNTRACK_CACHE,
    ZNTRACK_FIELD_DUMP,
    ZNTRACK_FIELD_LOAD,
    ZNTRACK_FIELD_SUFFIX,
    ZNTRACK_OPTION_PLOTS_CONFIG,
    FieldTypes,
)

# if t.TYPE_CHECKING:
from zntrack.node import Node
from zntrack.plugins import ZnTrackPlugin
from zntrack.utils import module_handler
from zntrack.utils.misc import (
    RunDVCImportPathHandler,
    get_attr_always_list,
    sort_and_deduplicate,
)
from zntrack.utils.node_wd import NWDReplaceHandler, nwd


@dataclasses.dataclass
class DVCPlugin(ZnTrackPlugin):
    def getter(self, field: dataclasses.Field) -> t.Any:
        getter = field.metadata.get(ZNTRACK_FIELD_LOAD)
        suffix = field.metadata.get(ZNTRACK_FIELD_SUFFIX)

        if getter is not None:
            if suffix is not None:
                return getter(self.node, field.name, suffix=suffix)
            return getter(self.node, field.name)
        return PLUGIN_EMPTY_RETRUN_VALUE

    def save(self, field: dataclasses.Field) -> None:
        dump_func = field.metadata.get(ZNTRACK_FIELD_DUMP)
        suffix = field.metadata.get(ZNTRACK_FIELD_SUFFIX)

        if dump_func is not None:
            if suffix is not None:
                dump_func(self.node, field.name, suffix=suffix)
            else:
                dump_func(self.node, field.name)

    def convert_to_params_yaml(self) -> dict | object:
        data = {}
        for field in dataclasses.fields(self.node):
            if field.metadata.get(FIELD_TYPE) == FieldTypes.PARAMS:
                data[field.name] = getattr(self.node, field.name)
            if field.metadata.get(FIELD_TYPE) == FieldTypes.DEPS:
                if getattr(self.node, field.name) is None:
                    continue
                content = getattr(self.node, field.name)
                if isinstance(content, (list, tuple, dict)):
                    new_content = []
                    for val in (
                        content
                        if isinstance(content, (list, tuple))
                        else content.values()
                    ):
                        if dataclasses.is_dataclass(val) and not isinstance(
                            val, (Node, znflow.Connection, znflow.CombinedConnections)
                        ):
                            # We save the values of the passed dataclasses
                            #  to the params.yaml file to be later used
                            #  by the DataclassContainer to recreate the
                            #  instance with the correct parameters.
                            dc_params = dataclasses.asdict(val)
                            dc_params["_cls"] = (
                                f"{module_handler(val)}.{val.__class__.__name__}"
                            )
                            new_content.append(dc_params)
                        elif isinstance(
                            val, (znflow.Connection, znflow.CombinedConnections)
                        ):
                            pass
                        else:
                            raise ValueError(
                                f"Found unsupported type '{type(val)}' ({val}) for DEPS"
                                f" field '{field.name}' in list"
                            )
                    if len(new_content) > 0:
                        data[field.name] = new_content
                elif dataclasses.is_dataclass(content) and not isinstance(
                    content, (Node, znflow.Connection, znflow.CombinedConnections)
                ):
                    dc_params = dataclasses.asdict(content)
                    dc_params["_cls"] = (
                        f"{module_handler(content)}.{content.__class__.__name__}"
                    )
                    data[field.name] = dc_params
                elif isinstance(content, (znflow.Connection, znflow.CombinedConnections)):
                    pass
                else:
                    raise ValueError(
                        f"Found unsupported type '{type(content)}' ({content})"
                        f" for DEPS field '{field.name}'"
                    )

        if len(data) > 0:
            return data
        return PLUGIN_EMPTY_RETRUN_VALUE

    def convert_to_dvc_yaml(self) -> dict | object:
        node_dict = converter.NodeConverter().encode(self.node)

        cmd = f"zntrack run {node_dict['module']}.{node_dict['cls']}"
        cmd += f" --name {node_dict['name']}"
        if hasattr(self.node, "_method"):
            cmd += f" --method {self.node._method}"
        stages = {
            "cmd": cmd,
            "metrics": [
                {
                    (self.node.nwd / "node-meta.json").as_posix(): {
                        "cache": config.ALWAYS_CACHE
                    }
                }
            ],
        }
        if self.node.always_changed:
            stages["always_changed"] = True
        plots = []

        nwd_handler = NWDReplaceHandler()

        for field in dataclasses.fields(self.node):
            if field.metadata.get(FIELD_TYPE) == FieldTypes.PARAMS:
                stages.setdefault(FieldTypes.PARAMS.value, []).append(self.node.name)
            elif field.metadata.get(FIELD_TYPE) == FieldTypes.PARAMS_PATH:
                if getattr(self.node, field.name) is None:
                    continue
                content = nwd_handler(
                    get_attr_always_list(self.node, field.name), nwd=self.node.nwd
                )
                content = [{pathlib.Path(x).as_posix(): None} for x in content]
                stages.setdefault(FieldTypes.PARAMS.value, []).extend(content)
            elif field.metadata.get(FIELD_TYPE) == FieldTypes.OUTS_PATH:
                if getattr(self.node, field.name) is None:
                    continue
                if getattr(self.node, field.name) == nwd:
                    raise ValueError(
                        "Can not use 'zntrack.nwd' directly as an output path. "
                        "Please use 'zntrack.nwd / <path/file>' instead."
                    )
                content = nwd_handler(
                    get_attr_always_list(self.node, field.name), nwd=self.node.nwd
                )
                content = [pathlib.Path(x).as_posix() for x in content]
                if field.metadata.get(ZNTRACK_CACHE) is False:
                    content = [{c: {"cache": False}} for c in content]
                stages.setdefault(FieldTypes.OUTS.value, []).extend(content)
            elif field.metadata.get(FIELD_TYPE) == FieldTypes.PLOTS_PATH:
                if getattr(self.node, field.name) is None:
                    continue
                content = nwd_handler(
                    get_attr_always_list(self.node, field.name), nwd=self.node.nwd
                )
                content = [pathlib.Path(x).as_posix() for x in content]
                if field.metadata.get(ZNTRACK_CACHE) is False:
                    content = [{c: {"cache": False}} for c in content]
                stages.setdefault(FieldTypes.OUTS.value, []).extend(content)
                # plots[self.node.name] = None
            elif field.metadata.get(FIELD_TYPE) == FieldTypes.METRICS_PATH:
                if getattr(self.node, field.name) is None:
                    continue
                content = nwd_handler(
                    get_attr_always_list(self.node, field.name), nwd=self.node.nwd
                )
                content = [pathlib.Path(x).as_posix() for x in content]
                if field.metadata.get(ZNTRACK_CACHE) is False:
                    content = [{c: {"cache": False}} for c in content]
                stages.setdefault(FieldTypes.METRICS.value, []).extend(content)
            elif field.metadata.get(FIELD_TYPE) == FieldTypes.OUTS:
                suffix = field.metadata[ZNTRACK_FIELD_SUFFIX]
                content = [(self.node.nwd / field.name).with_suffix(suffix).as_posix()]
                if field.metadata.get(ZNTRACK_CACHE) is False:
                    content = [{c: {"cache": False}} for c in content]
                stages.setdefault(FieldTypes.OUTS.value, []).extend(content)
            elif field.metadata.get(FIELD_TYPE) == FieldTypes.PLOTS:
                suffix = field.metadata[ZNTRACK_FIELD_SUFFIX]
                content = [(self.node.nwd / field.name).with_suffix(suffix).as_posix()]
                if field.metadata.get(ZNTRACK_CACHE) is False:
                    content = [{c: {"cache": False}} for c in content]
                stages.setdefault(FieldTypes.OUTS.value, []).extend(content)
                if ZNTRACK_OPTION_PLOTS_CONFIG in field.metadata:
                    file_path = (
                        (self.node.nwd / field.name).with_suffix(suffix).as_posix()
                    )
                    plots_config = field.metadata[ZNTRACK_OPTION_PLOTS_CONFIG].copy()
                    if "x" not in plots_config or "y" not in plots_config:
                        raise ValueError(
                            "Both 'x' and 'y' must be specified in the plots_config."
                        )
                    if "x" in plots_config:
                        plots_config["x"] = {file_path: plots_config["x"]}
                    if isinstance(plots_config["y"], list):
                        for idx, y in enumerate(plots_config["y"]):
                            cfg = copy.deepcopy(plots_config)
                            cfg["y"] = {file_path: y}
                            plots.append({f"{self.node.name}_{field.name}_{idx}": cfg})
                    else:
                        if "y" in plots_config:
                            plots_config["y"] = {file_path: plots_config["y"]}
                        plots.append({f"{self.node.name}_{field.name}": plots_config})
            elif field.metadata.get(FIELD_TYPE) == FieldTypes.METRICS:
                suffix = field.metadata[ZNTRACK_FIELD_SUFFIX]
                content = [(self.node.nwd / field.name).with_suffix(suffix).as_posix()]
                if field.metadata.get(ZNTRACK_CACHE) is False:
                    content = [{c: {"cache": False}} for c in content]
                stages.setdefault(FieldTypes.METRICS.value, []).extend(content)
            elif field.metadata.get(FIELD_TYPE) == FieldTypes.DEPS:
                if getattr(self.node, field.name) is None:
                    continue
                content = get_attr_always_list(self.node, field.name)
                paths = []
                for con in content:
                    if isinstance(con, (znflow.Connection)):
                        if con.item is not None:
                            raise NotImplementedError(
                                "znflow.Connection getitem is not supported yet."
                            )
                        paths.extend(
                            converter.node_to_output_paths(con.instance, con.attribute)
                        )
                    elif isinstance(con, (znflow.CombinedConnections)):
                        for _con in con.connections:
                            if con.item is not None:
                                raise NotImplementedError(
                                    "znflow.Connection getitem is not supported yet."
                                )
                            paths.extend(
                                converter.node_to_output_paths(
                                    _con.instance, _con.attribute
                                )
                            )
                    elif dataclasses.is_dataclass(con) and not isinstance(con, Node):
                        # add node name to params.yaml
                        stages.setdefault(FieldTypes.PARAMS.value, []).append(
                            self.node.name
                        )
                    else:
                        raise ValueError("unsupoorted type")

                if len(paths) > 0:
                    stages.setdefault(FieldTypes.DEPS.value, []).extend(paths)
            elif field.metadata.get(FIELD_TYPE) == FieldTypes.DEPS_PATH:
                if getattr(self.node, field.name) is None:
                    continue
                content = [
                    pathlib.Path(c).as_posix()
                    for c in get_attr_always_list(self.node, field.name)
                ]
                RunDVCImportPathHandler()(self.node.__dict__.get(field.name))
                stages.setdefault(FieldTypes.DEPS.value, []).extend(content)

        for key in stages:
            if key in ["cmd", "always_changed"]:
                continue
            stages[key] = sort_and_deduplicate(stages[key])

        return {"stages": stages, "plots": plots}

    def convert_to_zntrack_json(self, graph) -> dict | object:
        data = {
            "nwd": self.node.nwd,
        }
        for field in dataclasses.fields(self.node):
            if field.metadata.get(FIELD_TYPE) in [
                FieldTypes.PARAMS_PATH,
                FieldTypes.DEPS_PATH,
                FieldTypes.OUTS_PATH,
                FieldTypes.PLOTS_PATH,
                FieldTypes.METRICS_PATH,
                FieldTypes.DEPS,
            ]:
                data[field.name] = self.node.__dict__[field.name]

        data = znjson.dumps(
            data,
            indent=4,
            cls=znjson.ZnEncoder.from_converters(
                [
                    converter.ConnectionConverter,
                    converter.NodeConverter,
                    converter.CombinedConnectionsConverter,
                    znjson.converter.PathlibConverter,
                    converter.DVCImportPathConverter,
                    converter.DataclassConverter,
                ],
                add_default=False,
            ),
        )
        return json.loads(data)



"""Test that the dataclass post_init works properly."""

import zntrack


class NodeWithPostInit(zntrack.Node):
    def __post_init__(self):
        self.value = 42

    def run(self):
        pass


def test_create_node_with_post_init_plain(proj_path):
    node = NodeWithPostInit()
    assert node.value == 42


def test_create_node_with_post_init_project(proj_path):
    project = zntrack.Project()
    with project:
        node = NodeWithPostInit()
    assert node.value == 42


def test_create_node_with_post_init_from_rev(proj_path):
    node = NodeWithPostInit.from_rev()
    assert node.value == 42


def test_run_node_with_post_init(proj_path):
    project = zntrack.Project()
    with project:
        node = NodeWithPostInit()
    project.run()
    assert node.value == 42


def test_repro_node_with_post_init(proj_path):
    project = zntrack.Project()
    with project:
        node = NodeWithPostInit()
    project.repro()
    assert node.value == 42
    assert NodeWithPostInit.from_rev().value == 42



""""""

from pathlib import Path

import pytest

import zntrack


class StageIO(zntrack.Node):
    deps: Path = zntrack.deps_path()
    param: int = zntrack.params(1)
    outs: Path = zntrack.outs_path(Path("calculation.txt"))

    def run(self):
        """Actual computation"""
        with open(self.deps, "r") as f:
            file_content = f.readlines()

        Path(self.outs).write_text("".join(file_content))


class StageAddition(zntrack.Node):
    n_1: int = zntrack.params()
    n_2: int = zntrack.params()

    sum: int = zntrack.outs()
    dif: int = zntrack.outs()

    outs: Path = zntrack.outs_path(Path("calculation.txt"))

    def run(self):
        """Actual computation"""
        self.sum = self.n_1 + self.n_2
        self.dif = self.n_1 - self.n_2

        Path(self.outs).write_text(f"{self.n_1} + {self.n_2} = {self.sum}")


@pytest.mark.xfail(reason="Experiments not yet supported", strict=True)
def test_stage_addition(tmp_path_2):
    """Check that the dvc repro works"""
    project = zntrack.Project()

    with project:
        node = StageAddition(5, 10)

    project.repro()

    with project.create_experiment(name="exp1"):
        node

    with project.create_experiment(name="exp2"):
        node.n_1 = 50
        node.n_2 = 100

    project.run_exp()

    assert node.from_rev(rev="exp1").sum == 15
    assert node.from_rev(rev="exp2").sum == 150


def test_stage_io(proj_path):
    deps = Path("test_example_01.py").resolve()
    assert deps.is_file()

    with zntrack.Project() as project:
        stage_io = StageIO(deps=deps, param=1, name="Node")

    assert stage_io.deps == deps
    assert stage_io.param == 1
    assert stage_io.name == "Node"
    project.build()

    assert stage_io.param == 1
    assert stage_io.deps == deps
    project.run()

    assert stage_io.outs.read_text().startswith('"""')



import dataclasses
import json
import typing as t
from pathlib import Path

import znfields
import znjson

from zntrack import config
from zntrack.config import (
    FIELD_TYPE,
    NOT_AVAILABLE,
    ZNTRACK_CACHE,
    ZNTRACK_FIELD_LOAD,
    ZNTRACK_FILE_PATH,
    ZNTRACK_INDEPENDENT_OUTPUT_TYPE,
    ZNTRACK_LAZY_VALUE,
    FieldTypes,
)

# if t.TYPE_CHECKING:
from zntrack.node import Node
from zntrack.plugins import plugin_getter
from zntrack.utils.misc import TempPathLoader
from zntrack.utils.node_wd import NWDReplaceHandler

FIELD_PATH_TYPE = t.Union[
    str, Path, t.List[t.Union[str, Path]], dataclasses._MISSING_TYPE
]


def _paths_getter(self: Node, name: str):
    # TODO: if self._external_: try looking into
    # external/self.uuid/...
    # this works for everything except node-meta.json because that
    # defines the uuid
    nwd_handler = NWDReplaceHandler()

    if name in self.__dict__ and self.__dict__[name] is not ZNTRACK_LAZY_VALUE:
        return nwd_handler(self.__dict__[name], nwd=self.nwd)
    try:
        with self.state.fs.open(ZNTRACK_FILE_PATH) as f:
            content = json.load(f)[self.name][name]
            content = znjson.loads(json.dumps(content))

            if self.state.tmp_path is not None:
                loader = TempPathLoader()
                loader(content, instance=self)

            content = nwd_handler(content, nwd=self.nwd)

            return content
    except FileNotFoundError:
        return NOT_AVAILABLE


def outs_path(
    default: FIELD_PATH_TYPE = dataclasses.MISSING,
    *,
    cache: bool = True,
    independent: bool = False,
    **kwargs,
):
    """Define output file path(s) for a node.

    Parameters
    ----------
    default : str|Path|list[str|Path], optional
        Default path(s) to output files.
    cache : bool, optional
        Whether to use the DVC cache for the field. Default is `True`.
    independent : bool, optional
        Whether the output is independent of the node's inputs. Default is `False`.

    Examples
    --------
    >>> import zntrack
    >>> from pathlib import Path
    >>> class MyNode(zntrack.Node):
    ...     outs_path: Path = zntrack.outs_path(zntrack.nwd / "output.txt")
    ...
    ...     def run(self) -> None: ...
    ...         '''Save output to self.outs_path.'''
    """
    kwargs["metadata"] = kwargs.get("metadata", {})
    kwargs["metadata"][FIELD_TYPE] = FieldTypes.OUTS_PATH
    kwargs["metadata"][ZNTRACK_CACHE] = cache
    kwargs["metadata"][ZNTRACK_INDEPENDENT_OUTPUT_TYPE] = independent
    kwargs["metadata"][ZNTRACK_FIELD_LOAD] = _paths_getter
    return znfields.field(default=default, getter=plugin_getter, **kwargs)


def params_path(default: FIELD_PATH_TYPE = dataclasses.MISSING, **kwargs):
    """Define input parameter file path(s).

    Parameters
    ----------
    default : str|Path|list[str|Path], optional
        Path to one or multiple parameter files.

    Examples
    --------
    >>> import zntrack
    >>> class MyNode(zntrack.Node):
    ...     params_path: str = zntrack.params_path(default="params.yaml")
    ...
    ...     def run(self) -> None: ...
    ...
    >>> a = MyNode()
    >>> a.params_path
    'params.yaml'
    >>> b = MyNode(params_path="params2.yaml")
    >>> b.params_path
    'params2.yaml'

    """
    kwargs["metadata"] = kwargs.get("metadata", {})
    kwargs["metadata"][FIELD_TYPE] = FieldTypes.PARAMS_PATH
    kwargs["metadata"][ZNTRACK_CACHE] = True  # TODO: remove?
    kwargs["metadata"][ZNTRACK_FIELD_LOAD] = _paths_getter
    return znfields.field(default=default, getter=plugin_getter, **kwargs)


def plots_path(
    default: FIELD_PATH_TYPE = dataclasses.MISSING,
    *,
    cache: bool = True,
    independent: bool = False,
    **kwargs,
):
    """Create a field that handles plots and figure paths.

    Parameters
    ----------
    default : str|Path|list[str|Path], optional
        Path to one or multiple plot files.
        See https://dvc.org/doc/user-guide/experiment-management/visualizing-plots
        for more information.
    cache : bool, optional
        Whether to use the DVC cache for the field.
    independent : bool, optional
        Set to true if the output of this field can be independent of the
        node's inputs. E.g. if a csv file is produced that contains indices
        it might not change if the inputs to the node change.
        In such a case subsequent nodes might not rerun if
        independent is kept as False.

    Examples
    --------
    >>> import zntrack
    >>> from pathlib import Path
    >>> class MyNode(zntrack.Node):
    ...     plots_path: Path = zntrack.plots_path(zntrack.nwd / "plots.png")
    ...
    ...     def run(self) -> None: ...
    ...         '''Save a figure to self.plots_path.'''
    """
    kwargs["metadata"] = kwargs.get("metadata", {})
    kwargs["metadata"][FIELD_TYPE] = FieldTypes.PLOTS_PATH
    kwargs["metadata"][ZNTRACK_CACHE] = cache
    kwargs["metadata"][ZNTRACK_INDEPENDENT_OUTPUT_TYPE] = independent
    kwargs["metadata"][ZNTRACK_FIELD_LOAD] = _paths_getter
    return znfields.field(default=default, getter=plugin_getter, **kwargs)


def metrics_path(
    default: FIELD_PATH_TYPE = dataclasses.MISSING,
    *,
    cache: bool | None = None,
    independent: bool = False,
    **kwargs,
):
    """Define metrics file path(s) for a node.

    Parameters
    ----------
    default : str|Path|list[str|Path], optional
        Path to one or multiple metrics files.
    cache : bool, optional
        Whether to use the DVC cache for the field. If `None`,
        defaults to `zntrack.config.ALWAYS_CACHE`.
    independent : bool, optional
        Whether the output is independent of the node's inputs. Default is `False`.

    Examples
    --------
    >>> import zntrack
    >>> from pathlib import Path
    >>> class MyNode(zntrack.Node):
    ...     metrics_path: Path = zntrack.metrics_path(zntrack.nwd / "metrics.json")
    ...
    ...     def run(self) -> None: ...
    ...         '''Save metrics to self.metrics_path.'''
    """
    if cache is None:
        cache = config.ALWAYS_CACHE
    kwargs["metadata"] = kwargs.get("metadata", {})
    kwargs["metadata"][FIELD_TYPE] = FieldTypes.METRICS_PATH
    kwargs["metadata"][ZNTRACK_CACHE] = cache
    kwargs["metadata"][ZNTRACK_INDEPENDENT_OUTPUT_TYPE] = independent
    kwargs["metadata"][ZNTRACK_FIELD_LOAD] = _paths_getter
    return znfields.field(default=default, getter=plugin_getter, **kwargs)


def deps_path(
    default: FIELD_PATH_TYPE = dataclasses.MISSING, *, cache: bool = True, **kwargs
):
    """Define dependency file path(s) for a node.

    Parameters
    ----------
    default : str|Path|list[str|Path], optional
        Path to one or multiple dependency files.
    cache : bool, optional
        Whether to use the DVC cache for the field. Default is `True`.

    Examples
    --------
    >>> import zntrack
    >>> class MyNode(zntrack.Node):
    ...     dependencies: str = zntrack.deps_path()
    ...
    ...     def run(self) -> None: ...
    ...
    ... a = MyNode(dependencies=["file1.txt", "file2.txt"])
    """
    kwargs["metadata"] = kwargs.get("metadata", {})
    kwargs["metadata"][FIELD_TYPE] = FieldTypes.DEPS_PATH
    kwargs["metadata"][ZNTRACK_CACHE] = cache
    kwargs["metadata"][ZNTRACK_FIELD_LOAD] = _paths_getter
    return znfields.field(default=default, getter=plugin_getter, **kwargs)



# an exception if one tries to access node -
# data from a node that has not been loaded yet.
class ZnTrackError(Exception):
    """Base class for exceptions in this module."""

    pass


class InvalidOptionError(ZnTrackError, AttributeError):
    """Raised when using an invalid ZnTrackOption for a task."""



"""Test for [NodeAttribute] as 'zn.deps'."""

import dvc.cli

import zntrack.examples


def test_list_deps(proj_path):
    with zntrack.Project() as project:
        node1 = zntrack.examples.ParamsToOuts(params=1, name="node1")
        node2 = zntrack.examples.ParamsToOuts(params=2, name="node2")

        data = [node1.outs, node2.outs]
        node3 = zntrack.examples.SumNodeAttributes(inputs=data, shift=10)

    project.build()

    assert node1.name == "node1"
    assert node2.name == "node2"
    assert node3.name == "SumNodeAttributes"

    assert dvc.cli.main(["repro"]) == 0

    node1 = node1.from_rev(name=node1.name)
    node2 = node2.from_rev(name=node2.name)
    node3 = node3.from_rev(name=node3.name)

    assert node1.outs == 1
    assert node2.outs == 2
    assert node3.output == 13



import json
import pathlib

import pandas as pd
import yaml

import zntrack


class MyNode(zntrack.Node):
    parameter: int = zntrack.params()
    parameter_path: str = zntrack.params_path()

    # deps are tested separately

    deps_path: str = zntrack.deps_path()
    deps2_paths: list[pathlib.Path] = zntrack.deps_path()
    outs_path: list[str] = zntrack.outs_path()
    metrics_path: pathlib.Path = zntrack.metrics_path()
    plots_path: list[pathlib.Path] = zntrack.plots_path()

    outs_path_no_cache: list[pathlib.Path] = zntrack.outs_path(cache=False)
    plots_path_no_cache: list[pathlib.Path | str] = zntrack.plots_path(cache=False)
    metrics_path_cache: pathlib.Path = zntrack.metrics_path(cache=True)

    outs: int = zntrack.outs()
    metrics: dict = zntrack.metrics()
    plots: pd.DataFrame = zntrack.plots(y="y")

    outs_no_cache: int = zntrack.outs(cache=False)
    metrics_cache: dict = zntrack.metrics(cache=True)
    plots_no_cache: pd.DataFrame = zntrack.plots(cache=False, y="y")


CWD = pathlib.Path(__file__).parent.resolve()


def test_basic(proj_path):
    with zntrack.Project() as project:
        _ = MyNode(
            parameter=1,
            parameter_path="parameter.yaml",
            deps_path="deps.yaml",
            deps2_paths=[pathlib.Path("deps2.yaml")],
            outs_path=["outs.yaml"],
            metrics_path=zntrack.nwd / "my_metrics.json",
            plots_path=[zntrack.nwd / "my_plots.csv"],
            outs_path_no_cache=[zntrack.nwd / "no-cache-outs.json"],
            plots_path_no_cache=[
                zntrack.nwd / "no-cache-plots.csv",
                "no-cache-plots.json",
            ],
            metrics_path_cache=zntrack.nwd / "metrics-cache.json",
        )

    project.build()

    assert json.loads((CWD / "zntrack_config" / "basic.json").read_text()) == json.loads(
        (proj_path / "zntrack.json").read_text()
    )
    assert yaml.safe_load(
        (CWD / "dvc_config" / "basic.yaml").read_text()
    ) == yaml.safe_load((proj_path / "dvc.yaml").read_text())
    assert (CWD / "params_config" / "basic.yaml").read_text() == (
        proj_path / "params.yaml"
    ).read_text()


if __name__ == "__main__":
    test_basic(None)



"""Test params such as always_changed, ..."""

import json
import pathlib

import yaml

import zntrack

CWD = pathlib.Path(__file__).parent.resolve()


class MyNode(zntrack.Node):
    """Some Node."""

    def run(self) -> None:
        pass


def test_node(proj_path):
    with zntrack.Project() as proj:
        MyNode(name="some-node", always_changed=True)

    proj.repro()

    assert json.loads(
        (CWD / "zntrack_config" / "meta_params.json").read_text()
    ) == json.loads((proj_path / "zntrack.json").read_text())
    assert yaml.safe_load(
        (CWD / "dvc_config" / "meta_params.yaml").read_text()
    ) == yaml.safe_load((proj_path / "dvc.yaml").read_text())
    assert (CWD / "params_config" / "meta_params.yaml").read_text() == (
        proj_path / "params.yaml"
    ).read_text()


if __name__ == "__main__":
    test_node("")



import pytest

from zntrack.utils.misc import sort_and_deduplicate


def test_mixed_types():
    data = [
        "b.yaml",
        "a.yaml",
        {"c.yaml": {"cache": False}},
    ]
    expected_result = [
        "a.yaml",
        "b.yaml",
        {"c.yaml": {"cache": False}},
    ]
    assert sort_and_deduplicate(data) == expected_result


def test_all_strings():
    data = ["z.yaml", "a.yaml", "b.yaml", "a.yaml"]
    expected_result = ["a.yaml", "b.yaml", "z.yaml"]
    assert sort_and_deduplicate(data) == expected_result


def test_all_dicts():
    data = [
        {"a.yaml": {"cache": False}},
        {"c.yaml": {"cache": True}},
        {"b.yaml": {"cache": True}},
        {"a.yaml": {"cache": False}},  # Duplicate
    ]
    expected_result = [
        {"a.yaml": {"cache": False}},
        {"b.yaml": {"cache": True}},
        {"c.yaml": {"cache": True}},
    ]
    assert sort_and_deduplicate(data) == expected_result


def test_empty():
    data = []
    expected_result = []
    assert sort_and_deduplicate(data) == expected_result


def test_different_parameter():
    data = [
        {"a.yaml": {"cache": False}},
        {"a.yaml": {"cache": True}},
    ]
    with pytest.raises(ValueError):
        sort_and_deduplicate(data)


def test_different_parameter_and_type():
    data = [
        "a.yaml",
        {"a.yaml": {"cache": True}},
    ]
    with pytest.raises(ValueError):
        sort_and_deduplicate(data)


def test_different_parameter_and_type_rev():
    data = [
        {"a.yaml": {"cache": True}},
        "a.yaml",
    ]
    with pytest.raises(ValueError):
        sort_and_deduplicate(data)



import json
import pathlib

import yaml

import zntrack
import zntrack.group

CWD = pathlib.Path(__file__).parent.resolve()


class MyNode(zntrack.Node):
    parameter: int = zntrack.params()

    def run(self) -> None:
        pass


def test_group_nodes(proj_path):
    project = zntrack.Project()

    with project:
        a = MyNode(parameter=1)

    with project.group("Grp") as grp:
        b = MyNode(parameter=2)
        c = MyNode(parameter=3)
        d = MyNode(parameter=4)

    assert b in grp

    with project.group("Grp", "B") as grp_b:
        e = MyNode(parameter=5)

    project.build()

    assert a.state.group is None
    assert isinstance(b.state.group, zntrack.group.Group)
    assert b.state.group.names == ("Grp",)

    assert b in b.state.group
    assert c in b.state.group
    assert d in b.state.group

    assert len(b.state.group) == 3

    assert b.state.group == c.state.group == d.state.group == grp
    assert e.state.group == grp_b

    assert len(e.state.group) == 1

    assert a not in b.state.group

    assert a.name == "MyNode"
    assert b.name == "Grp_MyNode"
    assert c.name == "Grp_MyNode_1"
    assert d.name == "Grp_MyNode_2"

    assert json.loads((CWD / "zntrack_config" / "groups.json").read_text()) == json.loads(
        (proj_path / "zntrack.json").read_text()
    )
    assert yaml.safe_load(
        (CWD / "dvc_config" / "groups.yaml").read_text()
    ) == yaml.safe_load((proj_path / "dvc.yaml").read_text())
    assert (CWD / "params_config" / "groups.yaml").read_text() == (
        proj_path / "params.yaml"
    ).read_text()


if __name__ == "__main__":
    test_group_nodes("")



import numpy as np

import zntrack


class NumpyToOuts(zntrack.Node):
    value: np.ndarray = zntrack.outs()

    def run(self):
        self.value = np.array([1, 2, 3])


def test_numpy_outs(proj_path):
    project = zntrack.Project()
    with project:
        n = NumpyToOuts()
    project.repro()

    assert n.value.tolist() == [1, 2, 3]



import json
import pathlib

import pandas as pd

import zntrack


class MyNode(zntrack.Node):
    parameter: int = zntrack.params()
    parameter_path: str = zntrack.params_path()

    # deps are tested separately

    deps_path: str = zntrack.deps_path()
    deps2_paths: list[pathlib.Path] = zntrack.deps_path()
    outs_path: list[str] = zntrack.outs_path()
    metrics_path: pathlib.Path = zntrack.metrics_path()
    plots_path: list[pathlib.Path] = zntrack.plots_path()

    outs_path_no_cache: list[pathlib.Path] = zntrack.outs_path(cache=False)
    plots_path_no_cache: list[pathlib.Path | str] = zntrack.plots_path(cache=False)
    metrics_path_cache: pathlib.Path = zntrack.metrics_path(cache=True)

    outs: int = zntrack.outs()
    metrics: dict = zntrack.metrics()
    plots: pd.DataFrame = zntrack.plots(y="y")

    outs_no_cache: int = zntrack.outs(cache=False)
    metrics_cache: dict = zntrack.metrics(cache=True)
    plots_no_cache: pd.DataFrame = zntrack.plots(cache=False, y="y")

    def run(self):
        for path in self.deps2_paths:
            with open(path, "w") as f:
                f.write("Lorem ipsum")
        self.metrics_path.write_text(json.dumps({"a": 1}))
        df = pd.DataFrame({"y": [1, 2, 3]})
        for path in self.plots_path:
            df.to_csv(path)
        for path in self.outs_path:
            with open(path, "w") as f:
                f.write("Lorem ipsum")
        for path in self.outs_path_no_cache:
            with open(path, "w") as f:
                f.write("Lorem ipsum")
        for path in self.plots_path_no_cache:
            df.to_csv(path)
        self.metrics_path_cache.write_text(json.dumps({"a": 1}))

        self.outs = 42
        self.metrics = {"a": 1}
        self.plots = df

        self.outs_no_cache = 42
        self.metrics_cache = {"a": 1}
        self.plots_no_cache = df


def test_repro_basic(proj_path):
    with open("parameter.yaml", "w") as f:
        f.write("parameter: 1")
    with open("deps.yaml", "w") as f:
        f.write("deps: 1")
    with open("deps2.yaml", "w") as f:
        f.write("deps2: 1")

    with zntrack.Project() as project:
        node = MyNode(
            parameter=1,
            parameter_path="parameter.yaml",
            deps_path="deps.yaml",
            deps2_paths=[pathlib.Path("deps2.yaml")],
            outs_path=["outs.yaml"],
            metrics_path=zntrack.nwd / "my_metrics.json",
            plots_path=[zntrack.nwd / "my_plots.csv"],
            outs_path_no_cache=[zntrack.nwd / "no-cache-outs.json"],
            plots_path_no_cache=[
                zntrack.nwd / "no-cache-plots.csv",
                "no-cache-plots.csv",
            ],
            metrics_path_cache=zntrack.nwd / "metrics-cache.json",
        )

    project.repro()

    assert node.nwd == pathlib.Path("nodes/MyNode")

    assert pathlib.Path("outs.yaml").read_text() == "Lorem ipsum"
    assert (node.nwd / "no-cache-outs.json").read_text() == "Lorem ipsum"

    assert json.loads((node.nwd / "my_metrics.json").read_text()) == {"a": 1}
    assert json.loads((node.nwd / "metrics-cache.json").read_text()) == {"a": 1}

    assert pd.read_csv(node.nwd / "my_plots.csv", index_col=0).equals(
        pd.DataFrame({"y": [1, 2, 3]})
    )
    assert pd.read_csv("no-cache-plots.csv", index_col=0).equals(
        pd.DataFrame({"y": [1, 2, 3]})
    )

    assert json.loads((node.nwd / "outs.json").read_text()) == 42
    assert json.loads((node.nwd / "metrics.json").read_text()) == {"a": 1}
    assert pd.read_csv(node.nwd / "plots.csv", index_col=0).equals(
        pd.DataFrame({"y": [1, 2, 3]})
    )

    assert json.loads((node.nwd / "outs_no_cache.json").read_text()) == 42
    assert json.loads((node.nwd / "metrics_cache.json").read_text()) == {"a": 1}
    assert pd.read_csv(node.nwd / "plots_no_cache.csv", index_col=0).equals(
        pd.DataFrame({"y": [1, 2, 3]})
    )



import json

import znjson

from zntrack import config
from zntrack.config import NOT_AVAILABLE, FieldTypes
from zntrack.fields.base import field
from zntrack.node import Node


def _outs_getter(self: "Node", name: str, suffix: str):
    with self.state.fs.open((self.nwd / name).with_suffix(suffix)) as f:
        return json.load(f, cls=znjson.ZnDecoder)


def _outs_save_func(self: "Node", name: str, suffix: str):
    self.nwd.mkdir(parents=True, exist_ok=True)
    try:
        (self.nwd / name).with_suffix(suffix).write_text(
            znjson.dumps(getattr(self, name))
        )
    except TypeError as err:
        raise TypeError(f"Error while saving {name} to {self.nwd / name}.json") from err


def _metrics_save_func(self: "Node", name: str, suffix: str):
    self.nwd.mkdir(parents=True, exist_ok=True)
    try:
        (self.nwd / name).with_suffix(suffix).write_text(json.dumps(getattr(self, name)))
    except TypeError as err:
        raise TypeError(f"Error while saving {name} to {self.nwd / name}.json") from err


def outs(*, cache: bool = True, independent: bool = False, **kwargs):
    """Define output for a node.

    An output can be anything that can be pickled.

    Parameters
    ----------
    cache : bool, optional
       Set to true to use the DVC cache for the field.
       Default is ``zntrack.config.ALWAYS_CACHE``.
    independent : bool, optional
         Whether the output is independent of the node's inputs. Default is `False`.

    Examples
    --------
    >>> import zntrack
    >>> class MyNode(zntrack.Node):
    ...     outs: int = zntrack.outs()
    ...
    ...     def run(self) -> None:
    ...         '''Save output to self.outs.'''
    """
    return field(
        default=NOT_AVAILABLE,
        cache=cache,
        independent=independent,
        field_type=FieldTypes.OUTS,
        dump_fn=_outs_save_func,
        suffix=".json",
        load_fn=_outs_getter,
        repr=False,
        init=False,
        **kwargs,
    )


def metrics(*, cache: bool | None = None, independent: bool = False, **kwargs):
    """Define metrics for a node.

    The metrics must be a dictionary that can be serialized to JSON.

    Parameters
    ----------
    cache : bool, optional
       Set to true to use the DVC cache for the field.
       Default is ``zntrack.config.ALWAYS_CACHE``.
    independent : bool, optional
         Whether the output is independent of the node's inputs. Default is `False`.

    Examples
    --------
    >>> import zntrack
    >>> class MyNode(zntrack.Node):
    ...     metrics: dict = zntrack.metrics()
    ...
    ...     def run(self) -> None:
    ...         '''Save metrics to self.metrics.'''
    """
    if cache is None:
        cache = config.ALWAYS_CACHE
    return field(
        default=NOT_AVAILABLE,
        cache=cache,
        independent=independent,
        field_type=FieldTypes.METRICS,
        dump_fn=_metrics_save_func,
        suffix=".json",
        load_fn=_outs_getter,
        repr=False,
        init=False,
        **kwargs,
    )



"""Test correct stage/node names."""

import json
import pathlib

import yaml

import zntrack.examples

CWD = pathlib.Path(__file__).parent.resolve()


def test_names(proj_path):
    project = zntrack.Project()

    with project:
        zntrack.examples.ParamsToMetrics(params={"loss": 0.01})
        zntrack.examples.ParamsToMetrics(params={"loss": 0.02})
        zntrack.examples.ParamsToMetrics(params={"loss": 0.03}, name="Thanasis")

    with project.group("grp"):
        zntrack.examples.ParamsToMetrics(params={"loss": 0.04})
        zntrack.examples.ParamsToMetrics(params={"loss": 0.05})
        zntrack.examples.ParamsToMetrics(params={"loss": 0.06}, name="Cydney")

    with project.group("grp", "subgrp"):
        zntrack.examples.ParamsToMetrics(params={"loss": 0.07})
        zntrack.examples.ParamsToMetrics(params={"loss": 0.08})
        zntrack.examples.ParamsToMetrics(params={"loss": 0.09}, name="Lorine")

    project.build()

    assert json.loads((CWD / "zntrack_config" / "names.json").read_text()) == json.loads(
        (proj_path / "zntrack.json").read_text()
    )
    assert yaml.safe_load(
        (CWD / "dvc_config" / "names.yaml").read_text()
    ) == yaml.safe_load((proj_path / "dvc.yaml").read_text())
    assert (CWD / "params_config" / "names.yaml").read_text() == (
        proj_path / "params.yaml"
    ).read_text()



"""Test the apply function."""

import pytest

import zntrack.examples


@pytest.mark.parametrize("eager", [True, False])
def test_apply(proj_path, eager) -> None:
    """Test the "zntrack.apply" function."""
    project = zntrack.Project()

    JoinedParamsToOuts = zntrack.apply(zntrack.examples.ParamsToOuts, "join")  # noqa N806

    with project:
        a = zntrack.examples.ParamsToOuts(params=["a", "b"])
        b = JoinedParamsToOuts(params=["a", "b"])
        c = zntrack.apply(zntrack.examples.ParamsToOuts, "join")(params=["a", "b", "c"])

    if eager:
        project.run()
    else:
        project.repro(build=True)

    assert a.outs == ["a", "b"]
    assert b.outs == "a-b"
    assert c.outs == "a-b-c"


@pytest.mark.parametrize("attribute", [True, False])
@pytest.mark.parametrize("eager", [True, False])
def test_deps_apply(proj_path, eager, attribute):
    """Test connecting applied nodes to other nodes."""
    project = zntrack.Project()

    JoinedParamsToOuts = zntrack.apply(zntrack.examples.ParamsToOuts, "join")  # noqa N806

    assert issubclass(JoinedParamsToOuts, zntrack.Node)

    with project:
        a = zntrack.examples.ParamsToOuts(params=["a", "b"])
        b = JoinedParamsToOuts(params=["a", "b"])
        c = zntrack.apply(zntrack.examples.ParamsToOuts, "join")(params=["a", "b", "c"])

        if attribute:
            x3 = zntrack.examples.AddNodeAttributes(a=b.outs, b=c.outs)
        else:
            x3 = zntrack.examples.AddNodes2(a=b, b=c)

    if eager:
        project.run()
    else:
        project.repro(build=True)

    assert isinstance(a, zntrack.Node)
    assert isinstance(b, zntrack.Node)
    assert isinstance(c, zntrack.Node)
    assert isinstance(x3, zntrack.Node)

    assert x3.c == "a-ba-b-c"



import importlib.metadata
import logging
import sys

from zntrack import config
from zntrack.add import add
from zntrack.apply import apply
from zntrack.config import NOT_AVAILABLE, FieldTypes
from zntrack.fields import (
    deps,
    deps_path,
    field,
    metrics,
    metrics_path,
    outs,
    outs_path,
    params,
    params_path,
    plots,
    plots_path,
)
from zntrack.from_rev import from_rev
from zntrack.node import Node
from zntrack.project import Project
from zntrack.utils import nwd

__all__ = [
    "params",
    "deps",
    "outs",
    "plots",
    "metrics",
    "params_path",
    "deps_path",
    "outs_path",
    "plots_path",
    "metrics_path",
    "Node",
    "Project",
    "nwd",
    "from_rev",
    "apply",
    "add",
    "field",
    "FieldTypes",
    "NOT_AVAILABLE",
    "config",
]

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# Formatter for advanced logging
formatter = logging.Formatter("%(asctime)s - %(levelname)s: %(message)s")

channel = logging.StreamHandler(sys.stdout)
channel.setLevel(logging.INFO)
channel.setFormatter(formatter)

logger.addHandler(channel)

__version__ = importlib.metadata.version("zntrack")



import dataclasses
import functools
import os
import pathlib
import subprocess
import typing as t

import znfields

import zntrack
from zntrack.config import (
    FIELD_TYPE,
    NOT_AVAILABLE,
    PLUGIN_EMPTY_RETRUN_VALUE,
    ZNTRACK_CACHE,
)
from zntrack.plugins import ZnTrackPlugin, base_getter, plugin_getter


def _text_getter(self: zntrack.Node, name: str):
    with self.state.fs.open(self.nwd / f"{name}.txt", mode="r") as f:
        return f.read()


class TextPlugin(ZnTrackPlugin):
    def getter(self, field: dataclasses.Field) -> t.Any:
        if field.metadata.get(FIELD_TYPE) == "TextPlugin.text":
            return base_getter(self.node, field.name, _text_getter)

        return PLUGIN_EMPTY_RETRUN_VALUE

    def save(self, field: dataclasses.Field) -> None:
        if field.metadata.get(FIELD_TYPE) == "TextPlugin.text":
            if not pathlib.Path(self.node.nwd).exists():
                pathlib.Path(self.node.nwd).mkdir(parents=True)
            with open(self.node.nwd / f"{field.name}.txt", "w") as f:
                f.write(getattr(self.node, field.name))

    def convert_to_zntrack_json(self, graph=None):
        return PLUGIN_EMPTY_RETRUN_VALUE

    def convert_to_dvc_yaml(self):
        return PLUGIN_EMPTY_RETRUN_VALUE

    def convert_to_params_yaml(self):
        return PLUGIN_EMPTY_RETRUN_VALUE


@functools.wraps(znfields.field)
def text(*, cache: bool = True, **kwargs):
    kwargs["metadata"] = kwargs.get("metadata", {})
    kwargs["metadata"][FIELD_TYPE] = "TextPlugin.text"
    kwargs["metadata"][ZNTRACK_CACHE] = cache
    return znfields.field(default=NOT_AVAILABLE, getter=plugin_getter, **kwargs)


class TextNode(zntrack.Node):
    user: str = zntrack.params()

    comment: str = text()

    def run(self):
        self.comment = f"Hello, {self.user}!"


def test_simple_text(proj_path):
    os.environ["ZNTRACK_PLUGINS"] = (
        "tests.integration.test_options_custom.TextPlugin,zntrack.plugins.dvc_plugin.DVCPlugin"
    )
    try:
        with zntrack.Project() as proj:
            node = TextNode(user="Max")

        proj.build()
        subprocess.run(["dvc", "repro"], cwd=proj_path, check=True)

        with open(node.nwd / "comment.txt") as f:
            comment = f.read()

        assert comment == "Hello, Max!"

        assert node.comment == "Hello, Max!"

        node = node.from_rev()

        assert node.comment == "Hello, Max!"
    finally:
        del os.environ["ZNTRACK_PLUGINS"]



import numpy as np
import pytest
import yaml

import zntrack


class ComputeA(zntrack.Node):
    """Node stage A"""

    inp: np.ndarray = zntrack.params()
    out: np.ndarray = zntrack.outs()

    def run(self):
        self.out = np.power(2, self.inp)


def test_numpy_params(proj_path):
    """Check that the dvc repro works"""
    project = zntrack.Project()

    with project:
        ComputeA(inp=np.arange(5))

    with pytest.raises(yaml.representer.RepresenterError):
        project.repro()



import dataclasses
import json

import znflow
import znflow.handler
import znflow.utils
import znjson

from zntrack import converter
from zntrack.config import ZNTRACK_FILE_PATH, FieldTypes
from zntrack.fields.base import field
from zntrack.node import Node


def _deps_getter(self: "Node", name: str):
    with self.state.fs.open(ZNTRACK_FILE_PATH) as f:
        content = json.load(f)[self.name][name]
        # TODO: Ensure deps are loaded from the correct revision
        content = znjson.loads(
            json.dumps(content),
            cls=znjson.ZnDecoder.from_converters(
                [
                    converter.NodeConverter,
                    converter.ConnectionConverter,
                    converter.CombinedConnectionsConverter,
                    converter.DVCImportPathConverter,
                    converter.DataclassConverter,
                ],
                add_default=True,
            ),
        )
        if isinstance(content, converter.DataclassContainer):
            content = content.get_with_params(self.name, name)
        if isinstance(content, list):
            new_content = []
            idx = 0
            for val in content:
                if isinstance(val, converter.DataclassContainer):
                    new_content.append(val.get_with_params(self.name, name, idx))
                    idx += 1  # index only runs over dataclasses
                else:
                    new_content.append(val)
            content = new_content

        content = znflow.handler.UpdateConnectors()(content)

        return content


def deps(default=dataclasses.MISSING, **kwargs):
    """Define dependencies for a node.

    A Node dependency field can be used to pass data from one node to another.
    It can not be used to pass anything that is not a ``zntrack.Node`` or a ``dataclass``.

    Parameters
    ----------
    default : Any, optional
        Should not be set on the class level.

    Examples
    --------
    >>> import zntrack
    >>> class MyFirstNode(zntrack.Node):
    ...     outs: int = zntrack.outs()
    ...
    ...     def run(self) -> None:
    ...         self.outs = 42
    ...
    >>> class MySecondNode(zntrack.Node):
    ...     deps: int = zntrack.deps()
    ...
    ...     def run(self) -> None: ...
    ...
    >>> with zntrack.Project() as project:
    ...     node1 = MyFirstNode()
    ...     node2 = MySecondNode(deps=node1.outs)
    >>> project.build()
    """
    return field(
        default=default,
        load_fn=_deps_getter,
        field_type=FieldTypes.DEPS,
        **kwargs,
    )



import dataclasses
import pathlib
import subprocess
import typing as t

from znflow.node import NodeBaseMixin

PATH_LIKE = t.Union[str, pathlib.Path]


@dataclasses.dataclass(frozen=True)
class DVCImportPath(NodeBaseMixin):
    url: str
    path: pathlib.Path
    rev: str | None = None
    force: bool = False

    def run(self):
        if self.path.exists() and not self.force:
            dvc_file = self.path.parent / (self.path.name + ".dvc")
            if dvc_file.exists():
                return
            else:
                raise FileExistsError(f"{self.path} exists but no {dvc_file} file found")
        if self.rev:
            cmd = [
                "dvc",
                "import",
                self.url,
                self.path.as_posix(),
                "--rev",
                self.rev,
                "--no-exec",
            ]
            if self.force:
                cmd.append("--force")
            subprocess.check_call(cmd)
        else:
            cmd = ["dvc", "import-url", self.url, self.path.as_posix(), "--no-exec"]
            if self.force:
                cmd.append("--force")
            subprocess.check_call(cmd)


def add(
    url: str, path: str, rev: t.Optional[str] = None, force: bool = False
) -> DVCImportPath:
    return DVCImportPath(url, pathlib.Path(path), rev, force=force)



import dataclasses
import os
import pathlib

import git
import numpy.testing as npt
import pandas as pd
import pytest
import yaml

import zntrack.examples

try:
    import aim
except ImportError:
    pytest.skip("aim is not installed", allow_module_level=True)


@dataclasses.dataclass
class T1:
    temperature: float


@dataclasses.dataclass
class T2:
    temperature: float


class MD(zntrack.Node):
    t: T1 | T2 | list = zntrack.deps()

    result: str = zntrack.outs()

    def run(self):
        self.result = self.t.__class__.__name__


class RangePlotter(zntrack.Node):
    start: int = zntrack.params()
    stop: int = zntrack.params()

    plots: pd.DataFrame = zntrack.plots(y="range")

    def run(self):
        self.plots = pd.DataFrame({"idx": list(range(self.start, self.stop))})


# fixture to set the os.env before the test and remove if after the test
@pytest.fixture
def aim_proj_path(proj_path):
    os.environ["ZNTRACK_PLUGINS"] = (
        "zntrack.plugins.dvc_plugin.DVCPlugin,zntrack.plugins.aim_plugin.AIMPlugin"
    )
    os.environ["AIM_TRACKING_URI"] = "aim://127.0.0.1:43800"

    config = {
        "global": {
            "ZNTRACK_PLUGINS": os.environ["ZNTRACK_PLUGINS"],
            "AIM_TRACKING_URI": os.environ["AIM_TRACKING_URI"],
        }
    }
    pathlib.Path("env.yaml").write_text(yaml.dump(config))

    yield proj_path

    del os.environ["ZNTRACK_PLUGINS"]
    del os.environ["AIM_TRACKING_URI"]


def test_aim_metrics(aim_proj_path):
    proj = zntrack.Project()

    with proj:
        node = zntrack.examples.ParamsToMetrics(params={"loss": 0})

    proj.build()
    # there should be no entry in the mlflow server

    proj.repro(build=False)
    # # the run should be there

    with node.state.plugins["AIMPlugin"].get_aim_run() as run:
        df = run.dataframe()
        assert df["dvc_stage_name"].tolist() == ["ParamsToMetrics"]
        assert df["dvc_stage_hash"].tolist() == [node.state.get_stage_hash()]
        assert df["zntrack_node"].tolist() == ["zntrack.examples.nodes.ParamsToMetrics"]
        assert df["params.loss"].tolist() == [0]

        # metrics
        metrics = {}
        for metric in run.metrics():
            metrics[metric.name] = list(metric.data.values())[0]

        npt.assert_array_equal(metrics["metrics.loss"], [[0]])

    # make a git commit with all the changes
    proj.finalize(msg="test")
    repo = git.Repo()

    with node.state.plugins["AIMPlugin"].get_aim_run() as run:
        df = run.dataframe()
        assert df["dvc_stage_name"].tolist() == ["ParamsToMetrics"]
        assert df["dvc_stage_hash"].tolist() == [node.state.get_stage_hash()]
        assert df["zntrack_node"].tolist() == ["zntrack.examples.nodes.ParamsToMetrics"]
        assert df["params.loss"].tolist() == [0]
        assert df["git_commit_message"].tolist() == ["test"]
        assert df["git_commit_hash"].tolist() == [repo.head.commit.hexsha]


def test_aim_plotting(aim_proj_path):
    proj = zntrack.Project()

    with proj:
        node = RangePlotter(start=0, stop=10)

    proj.build()
    proj.repro(build=False)

    with node.state.plugins["AIMPlugin"].get_aim_run() as run:
        df = run.dataframe()
        assert df["dvc_stage_name"].tolist() == ["RangePlotter"]
        assert df["dvc_stage_hash"].tolist() == [node.state.get_stage_hash()]
        assert df["zntrack_node"].tolist() == ["test_plugins_aim.RangePlotter"]

        # metrics
        metrics = {}
        for metric in run.metrics():
            metrics[metric.name] = list(metric.data.values())[0]
        npt.assert_array_equal(metrics["plots.idx"], [list(range(10))])

    proj.finalize(msg="test")
    repo = git.Repo()

    with node.state.plugins["AIMPlugin"].get_aim_run() as run:
        df = run.dataframe()
        assert df["git_commit_message"].tolist() == ["test"]
        assert df["git_commit_hash"].tolist() == [repo.head.commit.hexsha]


def test_multiple_nodes(aim_proj_path):
    with zntrack.Project() as proj:
        a = zntrack.examples.ParamsToOuts(params=3)
        b = zntrack.examples.ParamsToOuts(params=7)
        c = zntrack.examples.SumNodeAttributesToMetrics(inputs=[a.outs, b.outs], shift=0)

    proj.repro()

    with c.state.plugins["AIMPlugin"].get_aim_run() as run:
        metrics = {}
        for metric in run.metrics():
            metrics[metric.name] = list(metric.data.values())[0]

        npt.assert_array_equal(metrics["metrics.value"], [[10.0]])

    with b.state.plugins["AIMPlugin"].get_aim_run() as run:
        b_run_id = run.hash

    proj.finalize(msg="exp1")
    repo = git.Repo()

    a.params = 5
    proj.repro()

    proj.finalize(msg="exp2")
    repo = git.Repo()

    a = a.from_rev(a.name)
    b = b.from_rev(b.name)
    c = c.from_rev(c.name)

    aim_repo = aim.Repo(path=os.environ["AIM_TRACKING_URI"])
    for run_metrics_col in aim_repo.query_metrics(
        (
            f"run.dvc_stage_name == '{a.name}' and "
            f"run.git_commit_hash == '{repo.head.commit.hexsha}'"
        )
    ).iter():
        assert "original_run_id" not in run_metrics_col.run.dataframe().columns

    for run_metrics_col in aim_repo.query_metrics(
        (
            f"run.dvc_stage_name == '{c.name}' and "
            f"run.git_commit_hash == '{repo.head.commit.hexsha}'"
        )
    ).iter():
        assert "original_run_id" not in run_metrics_col.run.dataframe().columns

    for run_metrics_col in aim_repo.query_metrics(
        (
            f"run.dvc_stage_name == '{b.name}' and "
            f"run.git_commit_hash == '{repo.head.commit.hexsha}'"
        )
    ).iter():
        assert run_metrics_col.run.dataframe()["original_run_id"].tolist() == [b_run_id]


def test_project_tags(aim_proj_path):
    with zntrack.Project(tags={"lorem": "ipsum", "hello": "world"}) as proj:
        a = zntrack.examples.ParamsToOuts(params=3)
        b = zntrack.examples.ParamsToOuts(params=7)
        zntrack.examples.SumNodeAttributesToMetrics(inputs=[a.outs, b.outs], shift=0)

    proj.repro()

    with a.state.plugins["AIMPlugin"].get_aim_run() as run:
        run: aim.Run
        assert set(run.tags) == {"lorem=ipsum", "hello=world"}


def test_dataclass_deps(aim_proj_path):
    t1 = T1(temperature=1)
    t2 = T2(temperature=1)

    with zntrack.Project() as proj:
        md = MD(t=t1)

    proj.repro()

    with md.state.plugins["AIMPlugin"].get_aim_run() as run:
        df = run.dataframe()
        assert df["t.temperature"].tolist() == [1.0]
        assert df["t._cls"].tolist() == ["test_plugins_aim.T1"]

    proj.finalize(msg="test")
    repo = git.Repo()

    md.t = t2
    proj.repro()

    with md.from_rev().state.plugins["AIMPlugin"].get_aim_run() as run:
        df = run.dataframe()
        assert df["t.temperature"].tolist() == [1.0]
        assert df["t._cls"].tolist() == ["test_plugins_aim.T2"]

    repo = git.Repo()
    repo.git.add(".")
    repo.git.commit("-m", "test")
    proj.finalize()

    md.t = [t1, t2]
    proj.repro()

    with md.from_rev().state.plugins["AIMPlugin"].get_aim_run() as run:
        df = run.dataframe()
        assert df["t"].tolist() == [
            (
                '[{"_cls": "test_plugins_aim.T1", "temperature": 1},'
                ' {"_cls": "test_plugins_aim.T2", "temperature": 1}]'
            )
        ]



import pathlib

import zntrack
from zntrack import Node, Project


class ChangeParamsInRun(Node):
    param: str = zntrack.params()

    def run(self):
        self.param = "incorrect param"


def test_ChangeParamsInRun(proj_path):
    with Project() as proj:
        ChangeParamsInRun(param="correct param")
    proj.repro()

    assert ChangeParamsInRun.from_rev().param == "correct param"


class ChangeJsonInRun(Node):
    outs: pathlib.Path = zntrack.outs_path(pathlib.Path("correct_out.txt"))

    def run(self):
        # need to create the file because DVC will fail otherwise
        self.outs.write_text("Create Correct File")
        self.outs = pathlib.Path("incorrect_out.txt")


def test_ChangeJsonInRun(proj_path):
    with Project() as proj:
        ChangeJsonInRun()
    proj.repro()
    assert ChangeJsonInRun.from_rev().outs == pathlib.Path("correct_out.txt")


class WriteToOutsOutsideRun(Node):
    outs: str = zntrack.outs()

    # def __init__(self, outs=None, **kwargs):
    #     super().__init__(**kwargs)
    #     self.outs = outs

    def run(self):
        self.outs = "correct outs"



"""Collection of common fixtures

References
----------
https://docs.pytest.org/en/6.2.x/fixture.html#scope-sharing-fixtures-across-classes-modules-packages-or-session

"""

import os
import pathlib
import shutil

import dvc.cli
import git
import pytest


@pytest.fixture
def tmp_path_2(tmp_path, request) -> pathlib.Path:
    """Temporary directory with the test file copied to it"""
    shutil.copy(request.module.__file__, tmp_path)
    os.chdir(tmp_path)
    return tmp_path


@pytest.fixture
def proj_path(tmp_path, request) -> pathlib.Path:
    """Temporary directory for testing DVC calls

    Parameters
    ----------
    tmp_path
    request: https://docs.pytest.org/en/6.2.x/reference.html#std-fixture-request

    Returns
    -------
    path to temporary directory

    """
    shutil.copy(request.module.__file__, tmp_path)
    os.chdir(tmp_path)
    git.Repo.init()
    dvc.cli.main(["init"])

    return tmp_path


@pytest.fixture
def empty_path(tmp_path, request) -> pathlib.Path:
    """Create an empty directory at the given path.

    Parameters
    ----------
    tmp_path
        The path to the directory to create.

    Returns
    -------
    pathlib.Path
        The path to the created directory.

    """
    shutil.copy(request.module.__file__, tmp_path)
    os.chdir(tmp_path)
    return tmp_path



import json
import pathlib
from typing import Tuple

from dvc.api import DVCFileSystem

from zntrack.config import ZNTRACK_FILE_PATH


def get_groups(remote, rev) -> Tuple[dict, list]:
    """Get the group names and the nodes in each group from the remote.

    Arguments:
    ---------
    remote : str
        The remote to get the group names from.
    rev : str
        The revision to use.

    Returns:
    -------
    groups : dict
        A nested dictionary with the group names as keys and the nodes in each group as
        values. Contains "short-name -> long-name" if inside a group.
    node_names: list
        A list of all node names in the project.

    """
    fs = DVCFileSystem(url=remote, rev=rev)
    with fs.open(ZNTRACK_FILE_PATH) as f:
        config = json.load(f)

    true_groups = {}
    node_names = []

    def add_to_group(groups, grp_names, node_name):
        """Recursively add node_name into the correct nested group structure."""
        if not grp_names:
            return

        current_group = grp_names[0]

        # If this is the last level, add the node directly
        if len(grp_names) == 1:
            if current_group not in groups:
                groups[current_group] = []
            groups[current_group].append(node_name)
        else:
            # Ensure the current group contains a dictionary inside a list
            if current_group not in groups:
                groups[current_group] = [{}]
            elif not isinstance(groups[current_group][0], dict):
                groups[current_group].insert(0, {})

            add_to_group(groups[current_group][0], grp_names[1:], node_name)

    for node_name, node_config in config.items():
        nwd = pathlib.Path(node_config["nwd"]["value"])
        grp_names = nwd.parent.as_posix().split("/")[1:]

        if not grp_names:
            node_names.append(node_name)
            grp_names = ["nodes"]
        else:
            for grp_name in grp_names:
                node_name = node_name.replace(f"{grp_name}_", "")

            node_names.append(f"{'_'.join(grp_names)}_{node_name}")
            node_name = f"{node_name} -> {node_names[-1]}"

        add_to_group(true_groups, grp_names, node_name)

    return true_groups, node_names



import zntrack.examples


def test_run_project(proj_path):
    with zntrack.Project() as project:
        node = zntrack.examples.ParamsToOuts(params=42)

    assert "outs" not in node.__dict__
    project.build()
    project.run()
    assert node.__dict__["outs"] == 42
    node = node.from_rev()
    assert "outs" not in node.__dict__
    node = node.from_rev(lazy_evaluation=False)
    assert node.__dict__["outs"] == 42


# def test_ParamsToOuts(proj_path, lazy, eager):
#         with zntrack.Project() as project:
#             node = zntrack.examples.ParamsToOuts(params=42)
#         project.run(eager=eager)

#         if not eager:
#             node.load()
#         if lazy and not eager:
#             assert node.__dict__["outs"] is ZNTRACK_LAZY_VALUE
#         else:
#             assert node.__dict__["outs"] is 42
#         assert node.outs == 42


# @pytest.mark.parametrize("eager", [True, False])
# @pytest.mark.parametrize("lazy", [True, False])
# def test_CollectOutputs(proj_path, lazy, eager):
#     with zntrack.config.updated_config(lazy=lazy):
#         with zntrack.Project() as project:
#             a = zntrack.examples.ParamsToOuts(params=17, name="a")
#             b = zntrack.examples.ParamsToOuts(params=42, name="b")
#             node = zntrack.examples.AddNodeNumbers(numbers=[a, b])
#         project.run(eager=eager)

#         if not eager:
#             node.load()
#         if lazy and not eager:
#             assert node.__dict__["sum"] is ZNTRACK_LAZY_VALUE
#             assert node.__dict__["numbers"] is ZNTRACK_LAZY_VALUEv
#             assert node.numbers[0].__dict__["outs"] is ZNTRACK_LAZY_VALUE
#             assert node.numbers[1].__dict__["outs"] is ZNTRACK_LAZY_VALUE
#         else:
#             assert node.__dict__["sum"] is 59
#             assert node.__dict__["numbers"][0].name == "a"
#             assert node.__dict__["numbers"][1].name == "b"

#         assert node.numbers[0].outs == 17
#         assert node.numbers[1].outs == 42
#         assert node.sum == 59

#         if not eager:
#             # Check that non-lazy loading works
#             node = node.from_rev(lazy=False)
#             assert node.__dict__["sum"] is 59
#             assert node.__dict__["numbers"][0].name == "a"
#             assert node.__dict__["numbers"][1].name == "b"



import dataclasses
import fnmatch
import json
import typing as t

import mlflow
import pandas as pd
import typer
import yaml
from dvc.api import DVCFileSystem
from mlflow.entities import Metric, Param, RunTag
from mlflow.tracking import MlflowClient
from mlflow.utils import mlflow_tags

from zntrack.cli.cli import app
from zntrack.config import FIELD_TYPE, FieldTypes
from zntrack.from_rev import from_rev
from zntrack.node import Node

numeric = t.Union[int, float]
metrics_type = dict[str, t.Union[numeric, list[numeric]]]
params_type = dict[str, t.Any]


@dataclasses.dataclass
class MLFlowNodeData:
    metrics: metrics_type
    params: params_type
    tags: dict[str, str]

    def upload(self, nested: bool = False) -> None:
        metrics: list[Metric] = []
        params: list[Param] = []
        tags: list[RunTag] = []

        for key, value in self.metrics.items():
            if isinstance(value, list):
                for i, v in enumerate(value):
                    metrics.append(Metric(key=key, value=v, step=i, timestamp=i))
            else:
                metrics.append(Metric(key=key, value=value, step=0, timestamp=0))

        for key, value in self.params.items():
            params.append(Param(key=key, value=str(value)))

        for key, value in self.tags.items():
            tags.append(RunTag(key=key, value=value))

        with mlflow.start_run(nested=nested) as active_run:
            mlflow_client = MlflowClient()
            mlflow_client.log_batch(
                run_id=active_run.info.run_id, metrics=metrics, params=params, tags=tags
            )

    @classmethod
    def from_node(cls, node: Node) -> "MLFlowNodeData":
        metrics = {}
        params = {}
        tags = {}

        for field in dataclasses.fields(node):
            if field.metadata.get(FIELD_TYPE) == FieldTypes.METRICS:
                for key, value in getattr(node, field.name).items():
                    metrics[f"{field.name}.{key}"] = value
            if field.metadata.get(FIELD_TYPE) == FieldTypes.PLOTS:
                df: pd.DataFrame = getattr(node, field.name)
                for column in df.columns:
                    metrics[f"{field.name}.{column}"] = df[column].values.tolist()
            if field.metadata.get(FIELD_TYPE) == FieldTypes.PARAMS:
                params[field.name] = getattr(node, field.name)
            # TODO: metrics_path, plots_path and params_path are currently being ignored
            # TODO: dataclass deps are also ignored

        tags[mlflow_tags.MLFLOW_RUN_NAME] = node.name
        if node.state.rev is not None:
            tags[mlflow_tags.MLFLOW_GIT_COMMIT] = node.state.rev
        if node.state.remote is not None:
            tags[mlflow_tags.MLFLOW_GIT_REPO_URL] = node.state.remote
        if hasattr(node, "__run_note__"):
            tags[mlflow_tags.MLFLOW_RUN_NOTE] = node.__run_note__()

        tags["dvc_stage_hash"] = node.state.get_stage_hash()
        tags["zntrack_node"] = f"{node.__module__}.{node.__class__.__name__}"

        return cls(metrics=metrics, params=params, tags=tags)


@app.command()
def mlflow_sync(
    nodes: list[str] | None = typer.Argument(
        None, help="ZnTrack nodes to sync. You can use glob patterns. (default: all)"
    ),
    rev: str | None = typer.Option(None, help="Git revision to load nodes from."),
    remote: str | None = typer.Option(None, help="Git remote to load nodes from."),
    experiment: str | None = typer.Option(
        None, help="MLFlow experiment name.", envvar="MLFLOW_EXPERIMENT_NAME"
    ),
    uri: str | None = typer.Option(
        None, help="MLFlow tracking URI.", envvar="MLFLOW_TRACKING_URI"
    ),
    parent: str | None = typer.Option(
        None,
        help="Specify a parent run name to group all nodes under.",
        envvar="MLFLOW_PARENT_RUN_NAME",
    ),
    dry: bool = typer.Option(
        False, "--dry", help="Print the data that would be uploaded."
    ),
) -> None:
    """Synchronize ZnTrack nodes with MLFlow."""
    fs = DVCFileSystem(url=remote, rev=rev)
    with fs.open("dvc.yaml", "r") as f:
        config = yaml.safe_load(f)

    node_lst: list[Node] = []

    for stage_name in config["stages"]:
        if nodes is not None:
            if not any(fnmatch.fnmatch(stage_name, node) for node in nodes):
                continue
        node_lst.append(from_rev(name=stage_name, rev=rev, remote=remote))

    data = []

    for node in node_lst:
        data.append(MLFlowNodeData.from_node(node))

    if dry:
        for node_data in data:
            print(json.dumps(dataclasses.asdict(node_data), indent=2))
        return

    if uri is not None:
        mlflow.set_tracking_uri(uri)

    if experiment is not None:
        mlflow.set_experiment(experiment)

    if parent is not None:
        with mlflow.start_run(run_name=parent):
            for node_data in data:
                node_data.upload(nested=True)
    else:
        for node_data in data:
            node_data.upload()



import pathlib
import typing as t

import pytest

import zntrack


class LoadFromDeps(zntrack.Node):
    data: t.Optional[t.Any] = zntrack.deps()
    file: t.Optional[pathlib.Path] = zntrack.deps_path()

    result: t.Any = zntrack.outs()

    def get_data(self):
        if self.data is not None:
            return self.data
        elif self.file is not None:
            return self.file.read_text()
        else:
            raise ValueError("No data provided")

    def run(self):
        self.result = self.get_data()


class WriteData(zntrack.Node):
    data: t.Any = zntrack.params()
    outs: t.Any = zntrack.outs()

    def run(self):
        self.outs = self.data


@pytest.mark.parametrize("value", ["data", "file"])
def test_init(value):
    with pytest.raises(TypeError):
        # setting only one of the deps args raise a requirement TypeError
        LoadFromDeps(**{value: None})


@pytest.mark.parametrize("eager", [True, False])
def test_from_zn_deps(proj_path, eager):
    with zntrack.Project() as proj:
        data = WriteData(data="Hello World")
        node = LoadFromDeps(data=data.outs, file=None)

    if eager:
        proj.run()
    else:
        proj.repro()

    assert node.result == "Hello World"


@pytest.mark.parametrize("eager", [True, False])
def test_from_dvc_deps(proj_path, eager):
    file = proj_path / "data.txt"
    file.write_text("Hello World")
    with zntrack.Project() as proj:
        node = LoadFromDeps(file=file, data=None)

    if eager:
        proj.run()
    else:
        proj.repro()

    assert node.result == "Hello World"


class EmptyNodesNode(zntrack.Node):
    # we use dvc.outs to generate zntrack.json
    file: pathlib.Path = zntrack.outs_path(zntrack.nwd / "file.txt")
    value: t.Optional[int] = zntrack.deps(None)
    outs: int = zntrack.outs()

    def run(self):
        if self.value is None:
            self.outs = 42
        else:
            self.outs = self.value
        self.file.write_text("Hello World")


@pytest.mark.parametrize("eager", [True, False])
def test_EmptyNode(proj_path, eager):
    with zntrack.Project() as project:
        node = EmptyNodesNode()
    if eager:
        project.run()
    else:
        project.repro()
    assert node.outs == 42



import pathlib

import zntrack.examples


def test_nodes_not_created(proj_path):
    project = zntrack.Project()

    with project:
        a = zntrack.examples.ParamsToOuts(
            params=42,
        )
        b = zntrack.examples.ParamsToOuts(
            params=18,
        )
        zntrack.examples.AddNodeAttributes(
            a=a.outs,
            b=b.outs,
        )

    with project.group("group"):
        d = zntrack.examples.ParamsToOuts(
            params=42,
        )
        e = zntrack.examples.ParamsToOuts(
            params=18,
        )
        zntrack.examples.AddNodeAttributes(
            a=d.outs,
            b=e.outs,
        )

    with project.group("plots"):
        zntrack.examples.WritePlots(
            x=[1, 2, 3],
            y=[4, 5, 6],
        )

    project.build()

    assert not pathlib.Path("nodes").exists()



import zntrack


class HelloWorld(zntrack.Node):
    """BasicTest class"""

    inputs: int = zntrack.params()
    output: int = zntrack.outs()

    def run(self):
        """Run method of the Node test instance"""
        self.output = self.inputs


def test_basic_io_assertion(proj_path):
    """Make a simple input/output assertion test for the nodes with different names"""
    project = zntrack.Project()
    with project:
        HelloWorld(inputs=3)
        HelloWorld(name="Test01", inputs=17)
        HelloWorld(name="Test02", inputs=42)

    project.run()

    assert HelloWorld.from_rev().output == 3
    assert HelloWorld.from_rev(name="Test01").output == 17
    assert HelloWorld.from_rev(name="Test02").output == 42



import pytest

import zntrack.examples
from zntrack.config import NodeStatusEnum


@pytest.mark.parametrize("eager", [True, False])
def test_AddNodes(proj_path, eager):
    with zntrack.Project() as project:
        add_numbers_a = zntrack.examples.AddNumbers(a=1, b=2, name="AddNumbersA")
        add_numbers_b = zntrack.examples.AddNumbers(a=1, b=3, name="AddNumbersB")
        add_nodes = zntrack.examples.AddNodes(a=add_numbers_a, b=add_numbers_b)

    assert add_numbers_a.state.state == NodeStatusEnum.CREATED
    assert add_numbers_b.state.state == NodeStatusEnum.CREATED
    assert add_nodes.state.state == NodeStatusEnum.CREATED

    if eager:
        project.run()
    else:
        project.repro()

    assert add_numbers_a.c == 3
    # TODO: Node status is not being updated when not using from_rev
    if eager:
        assert add_numbers_a.state.state == NodeStatusEnum.FINISHED
    assert add_numbers_b.c == 4
    if eager:
        assert add_numbers_b.state.state == NodeStatusEnum.FINISHED
    assert add_nodes.c == 7
    if eager:
        assert add_nodes.state.state == NodeStatusEnum.FINISHED


@pytest.mark.parametrize("eager", [True, False])
def test_AddNodeAttributes(proj_path, eager):
    with zntrack.Project() as project:
        add_numbers_a = zntrack.examples.AddNumbers(a=1, b=2, name="AddNumbersA")
        add_numbers_b = zntrack.examples.AddNumbers(a=1, b=3, name="AddNumbersB")
        add_nodes = zntrack.examples.AddNodeAttributes(
            a=add_numbers_a.c, b=add_numbers_b.c
        )

    if eager:
        project.run()
    else:
        project.repro()
    assert add_numbers_a.c == 3
    if eager:
        assert add_numbers_a.state.state == NodeStatusEnum.FINISHED
    assert add_numbers_b.c == 4
    if eager:
        assert add_numbers_b.state.state == NodeStatusEnum.FINISHED
    assert add_nodes.c == 7
    if eager:
        assert add_nodes.state.state == NodeStatusEnum.FINISHED


def test_AddNodes_legacy(proj_path):
    with zntrack.Project() as proj:
        add_numbers_a = zntrack.examples.AddNumbers(a=1, b=2, name="AddNumbersA")
        add_numbers_b = zntrack.examples.AddNumbers(a=1, b=3, name="AddNumbersB")
        add_nodes = zntrack.examples.AddNodeAttributes(
            a=add_numbers_a.c, b=add_numbers_b.c
        )

    proj.build()

    assert add_numbers_a.state.state == NodeStatusEnum.CREATED
    assert add_numbers_b.state.state == NodeStatusEnum.CREATED
    assert add_nodes.state.state == NodeStatusEnum.CREATED
    proj.repro(build=False)

    add_numbers_a = add_numbers_a.from_rev(name=add_numbers_a.name)
    add_numbers_b = add_numbers_b.from_rev(name=add_numbers_b.name)
    add_nodes = add_nodes.from_rev(name=add_nodes.name)

    assert add_numbers_a.c == 3
    assert add_numbers_a.state.state == NodeStatusEnum.FINISHED
    assert add_numbers_b.c == 4
    assert add_numbers_b.state.state == NodeStatusEnum.FINISHED
    assert add_nodes.c == 7
    assert add_nodes.state.state == NodeStatusEnum.FINISHED


def test_AddNodeAttributes_legacy(proj_path):
    with zntrack.Project() as proj:
        add_numbers_a = zntrack.examples.AddNumbers(a=1, b=2, name="AddNumbersA")
        add_numbers_b = zntrack.examples.AddNumbers(a=1, b=3, name="AddNumbersB")
        add_nodes = zntrack.examples.AddNodes(a=add_numbers_a, b=add_numbers_b)

    proj.build()

    assert add_numbers_a.state.state == NodeStatusEnum.CREATED
    assert add_numbers_b.state.state == NodeStatusEnum.CREATED
    assert add_nodes.state.state == NodeStatusEnum.CREATED

    proj.repro(build=False)

    add_numbers_a = add_numbers_a.from_rev(name=add_numbers_a.name)
    add_numbers_b = add_numbers_b.from_rev(name=add_numbers_b.name)
    add_nodes = add_nodes.from_rev(name=add_nodes.name)

    assert add_numbers_a.c == 3
    assert add_numbers_a.state.state == NodeStatusEnum.FINISHED
    assert add_numbers_b.c == 4
    assert add_numbers_b.state.state == NodeStatusEnum.FINISHED
    assert add_nodes.c == 7
    assert add_nodes.state.state == NodeStatusEnum.FINISHED


def test_OptionalDeps(proj_path):
    with zntrack.Project() as proj:
        add_numbers = zntrack.examples.AddNumbers(a=1, b=2)
        add_none = zntrack.examples.OptionalDeps()
        add_value = zntrack.examples.OptionalDeps(value=add_numbers.c)

    proj.run()

    assert add_numbers.c == 3
    assert add_none.result == 0.0
    assert add_value.result == 3.0



import pathlib
import typing as t

import pytest

import zntrack

# TODO UNCLEAR ERROR, __file__ attribute is not the
#  same as the test file we want to collect


class WriteToNWD(zntrack.Node):
    text: str = zntrack.params()
    file: tuple[pathlib.Path] = zntrack.outs_path((zntrack.nwd / "test.txt",))

    def run(self):
        self.nwd.mkdir(exist_ok=True, parents=True)
        self.file[0].write_text(self.text)


class FileToOuts(zntrack.Node):
    # although, this is a file path, it has to be zn.deps
    file: tuple[pathlib.Path] = zntrack.deps()
    text: str = zntrack.outs()

    def run(self):
        with open(self.file[0], "r") as f:
            self.text = f.read()


@pytest.mark.parametrize("eager", [True, False])
def test_WriteToNWD(proj_path, eager):
    with zntrack.Project() as project:
        write_to_nwd = WriteToNWD(text="Hello World")
        file_to_outs = FileToOuts(file=write_to_nwd.file)

    if eager:
        project.run()
    else:
        project.repro()

    assert write_to_nwd.file[0].read_text() == "Hello World"
    assert write_to_nwd.file == (pathlib.Path("nodes", "WriteToNWD", "test.txt"),)

    assert write_to_nwd.__dict__["file"] == (pathlib.Path("$nwd$", "test.txt"),)
    assert file_to_outs.text == "Hello World"


def test_OutAsNWD(proj_path):
    class OutsAsNWD(zntrack.Node):
        text: t.Any = zntrack.params()
        outs: pathlib.Path = zntrack.outs_path(zntrack.nwd)



import pathlib
import typing as t

import znflow

from zntrack.config import NWD_PATH

if t.TYPE_CHECKING:
    from zntrack import Node


def _extract_group_from_nwd(path: pathlib.Path) -> tuple | None:
    # Convert the path to a list of parts
    parts = list(path.parts)

    # Check if the path starts with 'nodes'
    if parts[0] != str(NWD_PATH):
        raise ValueError(f"Path must start with 'nodes', found {parts}")

    # Check the length of the path to determine the output
    if len(parts) == 2:  # no groups
        return None
    else:  # groups
        return tuple(parts[1:-1])


class Group:
    def __init__(self, names: tuple[str], nodes: list | None = None) -> None:
        self._names = names
        self._nodes = nodes if nodes else []

    @property
    def nwd(self) -> pathlib.Path:
        return NWD_PATH.joinpath(*self._names)

    @property
    def names(self):
        return self._names

    @property
    def nodes(self):
        return self._nodes

    def __eq__(self, value: "Group") -> bool:
        if not isinstance(value, self.__class__):
            return False
        return value.names == self.names

    def __repr__(self) -> str:
        return f"Group(name='{self._names}')"

    def __str__(self) -> str:
        return f"Group(name='{self._names}')"

    def __contains__(self, item: "Node|str") -> bool:
        """Check if the Node is in the group."""
        if isinstance(item, str):
            return item in [node.name for node in self._nodes]
        return item in self._nodes

    def __iter__(self) -> t.Iterator["Node"]:
        """Iterate over the nodes in the group."""
        return iter(self._nodes)

    def __getitem__(self, name: str) -> "Node":
        """Get the Node from the group."""
        for node in self._nodes:
            if node.name == name:
                return node
        raise KeyError(f"Node {name} not found.")

    def __len__(self) -> int:
        """Get the number of nodes in the group."""
        return len(self._nodes)

    @classmethod
    def from_znflow_group(cls, group: znflow.Group) -> "Group":
        return cls(names=group.names, nodes=group.nodes)

    @classmethod
    def from_nwd(cls, nwd: pathlib.Path) -> "Group|None":
        names = _extract_group_from_nwd(nwd)
        if names is None:
            return None
        return cls(names=names)



import dataclasses
import os
import pathlib
import uuid

import git
import mlflow
import pandas as pd
import pytest
import yaml
from mlflow.utils import mlflow_tags

import zntrack.examples


@dataclasses.dataclass
class T1:
    temperature: float


@dataclasses.dataclass
class T2:
    temperature: float


class MD(zntrack.Node):
    t: T1 | T2 | list = zntrack.deps()

    result: str = zntrack.outs()

    def run(self):
        self.result = self.t.__class__.__name__

    def __run_note__(self) -> str:
        return f"""
This is the MD Node from rev {self.state.rev}.

Run it via `MD(t=T1(temperature=1))`

within

```python
with zntrack.Project() as proj:
    md = MD(t=T1(temperature=1))
proj.repro()
```
"""


class RangePlotter(zntrack.Node):
    start: int = zntrack.params()
    stop: int = zntrack.params()

    plots: pd.DataFrame = zntrack.plots(y="range")

    def run(self):
        self.plots = pd.DataFrame({"idx": list(range(self.start, self.stop))})


@pytest.fixture
def mlflow_proj_path(proj_path):
    os.environ["ZNTRACK_PLUGINS"] = (
        "zntrack.plugins.dvc_plugin.DVCPlugin,zntrack.plugins.mlflow_plugin.MLFlowPlugin"
    )
    os.environ["MLFLOW_TRACKING_URI"] = "http://127.0.0.1:5000"
    os.environ["MLFLOW_EXPERIMENT_NAME"] = f"test-{uuid.uuid4()}"

    config = {
        "global": {
            "ZNTRACK_PLUGINS": os.environ["ZNTRACK_PLUGINS"],
            "MLFLOW_TRACKING_URI": os.environ["MLFLOW_TRACKING_URI"],
            "MLFLOW_EXPERIMENT_NAME": os.environ["MLFLOW_EXPERIMENT_NAME"],
        }
    }
    pathlib.Path("env.yaml").write_text(yaml.dump(config))

    yield proj_path

    del os.environ["ZNTRACK_PLUGINS"]
    del os.environ["MLFLOW_TRACKING_URI"]
    del os.environ["MLFLOW_EXPERIMENT_NAME"]


def test_mlflow_metrics(mlflow_proj_path):
    proj = zntrack.Project()

    with proj:
        node = zntrack.examples.ParamsToMetrics(params={"loss": 0})

    proj.build()
    # there should be no entry in the mlflow server

    proj.repro(build=False)
    # # the run should be there

    with node.state.plugins["MLFlowPlugin"]:
        pass  # load run_id states

    child_run_id = node.state.plugins["MLFlowPlugin"].child_run_id
    parent_run_id = node.state.plugins["MLFlowPlugin"].parent_run_id

    assert child_run_id is not None
    assert parent_run_id is not None

    run = mlflow.get_run(child_run_id)
    # assert params are logged
    assert run.data.params == {"params": "{'loss': 0}"}  # this is strange!
    # assert tags
    assert run.data.tags["dvc_stage_name"] == "ParamsToMetrics"
    assert run.data.tags["dvc_stage_hash"] == node.state.get_stage_hash()
    assert run.data.tags["zntrack_node"] == "zntrack.examples.nodes.ParamsToMetrics"

    # assert metrics
    assert run.data.metrics == {"metrics.loss": 0.0}

    # make a git commit with all the changes
    proj.finalize(msg="test")
    repo = git.Repo()

    run = mlflow.get_run(child_run_id)  # need to query the run again

    assert run.data.tags["git_commit_message"] == "test"
    assert run.data.tags["git_commit_hash"] == repo.head.commit.hexsha


def test_mlflow_plotting(mlflow_proj_path):
    proj = zntrack.Project()

    with proj:
        node = RangePlotter(start=0, stop=10)

    proj.build()
    proj.repro(build=False)

    with node.state.plugins["MLFlowPlugin"]:
        pass  # load run_id states

    child_run_id = node.state.plugins["MLFlowPlugin"].child_run_id
    parent_run_id = node.state.plugins["MLFlowPlugin"].parent_run_id

    assert child_run_id is not None
    assert parent_run_id is not None

    run = mlflow.get_run(child_run_id)
    # assert params are logged
    assert run.data.params == {"start": "0", "stop": "10"}
    # assert tags
    assert run.data.tags["dvc_stage_name"] == "RangePlotter"
    assert run.data.tags["dvc_stage_hash"] == node.state.get_stage_hash()
    assert run.data.tags["zntrack_node"] == "test_plugins_mlflow.RangePlotter"

    # assert metrics (last)
    assert run.data.metrics == {"plots.idx": 9.0}

    client = mlflow.MlflowClient()
    history = client.get_metric_history(child_run_id, "plots.idx")
    assert len(history) == 10
    assert [entry.value for entry in history] == list(range(10))

    # make a git commit with all the changes
    repo = git.Repo()
    repo.git.add(".")
    repo.git.commit("-m", "test")
    node.state.plugins["MLFlowPlugin"].finalize()

    run = mlflow.get_run(child_run_id)  # need to query the run again

    assert run.data.tags["git_commit_message"] == "test"
    assert run.data.tags["git_commit_hash"] == repo.head.commit.hexsha


@pytest.mark.parametrize("skip_cached", [True, False])
def test_multiple_nodes(mlflow_proj_path, skip_cached):
    with zntrack.Project() as proj:
        a = zntrack.examples.ParamsToOuts(params=3)
        b = zntrack.examples.ParamsToOuts(params=7)
        c = zntrack.examples.SumNodeAttributesToMetrics(inputs=[a.outs, b.outs], shift=0)

    proj.repro()

    assert c.metrics == {"value": 10.0}

    with a.state.plugins["MLFlowPlugin"]:
        mlflow.get_run(a.state.plugins["MLFlowPlugin"].child_run_id)

    with b.state.plugins["MLFlowPlugin"]:
        b_run = mlflow.get_run(b.state.plugins["MLFlowPlugin"].child_run_id)

    with c.state.plugins["MLFlowPlugin"]:
        c_run = mlflow.get_run(c.state.plugins["MLFlowPlugin"].child_run_id)

    assert c_run.data.metrics == {"metrics.value": 10.0}

    proj.finalize(msg="exp1", skip_cached=skip_cached, update_run_names=not skip_cached)
    repo = git.Repo()

    a.params = 5
    proj.repro()

    proj.finalize(msg="exp2", skip_cached=skip_cached, update_run_names=not skip_cached)
    repo = git.Repo()

    # find all runs with `git_commit_hash` == repo.head.commit.hexsha
    runs = mlflow.search_runs(
        filter_string=f"tags.git_commit_hash = '{repo.head.commit.hexsha}'"
    )
    if skip_cached:
        assert len(runs) == 3
    else:
        assert len(runs) == 4

    a_run_2 = mlflow.search_runs(
        filter_string=(
            f"tags.git_commit_hash = '{repo.head.commit.hexsha}' "
            f"and tags.dvc_stage_name = '{a.name}'"
        ),
        output_format="list",
    )
    assert len(a_run_2) == 1
    a_run_2 = a_run_2[0]

    b_run_2 = mlflow.search_runs(
        filter_string=(
            f"tags.git_commit_hash = '{repo.head.commit.hexsha}'"
            f" and tags.dvc_stage_name = '{b.name}'"
        ),
        output_format="list",
    )
    if skip_cached:
        assert len(b_run_2) == 0
    else:
        assert len(b_run_2) == 1
        b_run_2 = b_run_2[0]
        assert b_run_2.data.tags["original_run_id"] == b_run.info.run_id
        # original runs will not be updated with a new name to
        # indicate that they are cached
        assert b_run_2.data.tags[mlflow_tags.MLFLOW_RUN_NAME] == "ParamsToOuts_1"

    c_run_2 = mlflow.search_runs(
        filter_string=(
            f"tags.git_commit_hash = '{repo.head.commit.hexsha}'"
            f" and tags.dvc_stage_name = '{c.name}'"
        ),
        output_format="list",
    )
    assert len(c_run_2) == 1
    c_run_2 = c_run_2[0]

    assert "original_run_id" not in a_run_2.data.tags
    assert "original_run_id" not in c_run_2.data.tags

    if skip_cached:
        assert a_run_2.data.tags[mlflow_tags.MLFLOW_RUN_NAME] == "ParamsToOuts"
        assert (
            c_run_2.data.tags[mlflow_tags.MLFLOW_RUN_NAME] == "SumNodeAttributesToMetrics"
        )
    else:
        assert a_run_2.data.tags[mlflow_tags.MLFLOW_RUN_NAME] == "exp2:ParamsToOuts"
        assert (
            c_run_2.data.tags[mlflow_tags.MLFLOW_RUN_NAME]
            == "exp2:SumNodeAttributesToMetrics"
        )

    assert c_run_2.data.metrics == {"metrics.value": 12.0}


def test_project_tags(mlflow_proj_path):
    with zntrack.Project(tags={"lorem": "ipsum", "hello": "world"}) as proj:
        a = zntrack.examples.ParamsToOuts(params=3)
        b = zntrack.examples.ParamsToOuts(params=7)
        c = zntrack.examples.SumNodeAttributesToMetrics(inputs=[a.outs, b.outs], shift=0)

    proj.repro()

    with a.state.plugins["MLFlowPlugin"]:
        a_run = mlflow.get_run(a.state.plugins["MLFlowPlugin"].child_run_id)
        parent_run = mlflow.get_run(a.state.plugins["MLFlowPlugin"].parent_run_id)

    with b.state.plugins["MLFlowPlugin"]:
        b_run = mlflow.get_run(b.state.plugins["MLFlowPlugin"].child_run_id)

    with c.state.plugins["MLFlowPlugin"]:
        c_run = mlflow.get_run(c.state.plugins["MLFlowPlugin"].child_run_id)

    assert a_run.data.tags["lorem"] == "ipsum"
    assert a_run.data.tags["hello"] == "world"

    assert b_run.data.tags["lorem"] == "ipsum"
    assert b_run.data.tags["hello"] == "world"

    assert c_run.data.tags["lorem"] == "ipsum"
    assert c_run.data.tags["hello"] == "world"

    assert parent_run.data.tags["lorem"] == "ipsum"
    assert parent_run.data.tags["hello"] == "world"


def test_dataclass_deps(mlflow_proj_path):
    t1 = T1(temperature=1)
    t2 = T2(temperature=1)

    with zntrack.Project() as proj:
        md = MD(t=t1)

    proj.repro()

    with md.state.plugins["MLFlowPlugin"]:
        run = mlflow.get_run(md.state.plugins["MLFlowPlugin"].child_run_id)

    assert (
        run.data.params["t"] == "[{'temperature': 1, '_cls': 'test_plugins_mlflow.T1'}]"
    )

    proj.finalize(msg="run1 exp.")
    git.Repo()

    mdx = MD.from_rev()
    assert mdx.__run_note__() != ""
    with mdx.state.plugins["MLFlowPlugin"]:
        run = mlflow.get_run(mdx.state.plugins["MLFlowPlugin"].child_run_id)
    assert run.data.tags[mlflow_tags.MLFLOW_RUN_NAME] == "run1:MD"

    md.t = t2
    proj.repro()

    md = MD.from_rev()

    with md.state.plugins["MLFlowPlugin"]:
        run = mlflow.get_run(md.state.plugins["MLFlowPlugin"].child_run_id)

    assert (
        run.data.params["t"] == "[{'temperature': 1, '_cls': 'test_plugins_mlflow.T2'}]"
    )

    proj.finalize(msg="run2 exp.")

    mdx = MD.from_rev()
    with mdx.state.plugins["MLFlowPlugin"]:
        run = mlflow.get_run(mdx.state.plugins["MLFlowPlugin"].child_run_id)
    assert run.data.tags[mlflow_tags.MLFLOW_RUN_NAME] == "run2:MD"

    with zntrack.Project() as proj:
        md = MD(t=[t1, t2])
    proj.repro()

    md = MD.from_rev()

    with md.state.plugins["MLFlowPlugin"]:
        run = mlflow.get_run(md.state.plugins["MLFlowPlugin"].child_run_id)

    assert run.data.params["t"] == (
        "[{'temperature': 1, '_cls': 'test_plugins_mlflow.T1'},"
        " {'temperature': 1, '_cls': 'test_plugins_mlflow.T2'}]"
    )

    md = zntrack.from_rev(md.name)
    assert hasattr(md, "__run_note__")
    assert md.__run_note__() != ""
    # assert run.data.tags[mlflow_tags.MLFLOW_RUN_NAME] == "run2:MD"



"""Collection of ZnTrack example Nodes.

These nodes are primarily used for testing and demonstration purposes.
"""

import json
import pathlib
import random
import typing as t

import pandas as pd

import zntrack


class ReadFile(zntrack.Node):
    """Read a file."""

    path: pathlib.Path = zntrack.deps_path()
    content: str = zntrack.outs()

    def run(self):
        """Read a file."""
        self.content = self.path.read_text()


class ParamsToOuts(zntrack.Node):
    """Save params to outs."""

    params: t.Any = zntrack.params()
    outs: t.Any = zntrack.outs()

    def run(self) -> None:
        """Save params to outs."""
        self.outs = self.params

    def join(self) -> None:
        """Join the results."""
        self.outs = "-".join(self.params)


class ParamsToMetrics(zntrack.Node):
    """Save params to metrics."""

    params: dict = zntrack.params()
    metrics: dict = zntrack.metrics()

    def run(self) -> None:
        """Save params to metrics."""
        self.metrics = self.params

    def __run_note__(self) -> str:
        """Markdown style run note."""
        return "This is a test run note."


class DepsToMetrics(zntrack.Node):
    """Save params to metrics."""

    deps: dict = zntrack.deps()
    metrics: dict = zntrack.metrics()

    def run(self) -> None:
        """Save params to metrics."""
        self.metrics = self.deps


class WritePlots(zntrack.Node):
    """Generate a plot."""

    plots: pd.DataFrame = zntrack.plots(x="x", y="y")
    x: list = zntrack.params(default_factory=lambda: [1, 2, 3])
    y: list = zntrack.params(default_factory=lambda: [4, 5, 6])

    def run(self):
        """Write plots."""
        self.plots = pd.DataFrame({"x": self.x, "y": self.y})


class AddNumbers(zntrack.Node):
    """Add two numbers."""

    a: float = zntrack.params()
    b: float = zntrack.params()
    c: float = zntrack.outs()

    def run(self):
        """Add two numbers."""
        self.c = self.a + self.b


class AddNumbersProperty(zntrack.Node):
    """Add two numbers."""

    a: float = zntrack.params()
    b: float = zntrack.params()

    @property
    def c(self):
        """Add two numbers."""
        return self.a + self.b


class AddNodes(zntrack.Node):
    """Add two nodes."""

    a: AddNumbers = zntrack.deps()
    b: AddNumbers = zntrack.deps()
    c: float = zntrack.outs()

    def run(self):
        """Add two nodes."""
        self.c = self.a.c + self.b.c


class AddNodes2(zntrack.Node):
    """Add two nodes."""

    a: ParamsToOuts = zntrack.deps()
    b: ParamsToOuts = zntrack.deps()
    c: float = zntrack.outs()

    def run(self):
        """Add two nodes."""
        self.c = self.a.outs + self.b.outs


class AddNodeAttributes(zntrack.Node):
    """Add two node attributes."""

    a: float = zntrack.deps()
    b: float = zntrack.deps()
    c: float = zntrack.outs()

    def run(self):
        """Add two node attributes."""
        self.c = self.a + self.b


class AddNodeNumbers(zntrack.Node):
    """Add up all 'x.outs' from the dependencies."""

    numbers: list = zntrack.deps()
    sum: int = zntrack.outs()

    def run(self):
        """Add up all 'x.outs' from the dependencies."""
        self.sum = sum(x.outs for x in self.numbers)


class SumNodeAttributes(zntrack.Node):
    """Sum a list of numbers."""

    inputs: list[float | int] = zntrack.deps()
    shift: int | float = zntrack.params()
    output: int | float = zntrack.outs()

    def run(self) -> None:
        """Sum a list of numbers."""
        self.output = sum(self.inputs) + self.shift


class SumNodeAttributesToMetrics(zntrack.Node):
    """Sum a list of numbers."""

    inputs: list[float | int] = zntrack.deps()
    shift: int | float = zntrack.params()
    metrics: dict[str, int | float] = zntrack.metrics()

    def run(self) -> None:
        """Sum a list of numbers."""
        self.metrics = {"value": sum(self.inputs) + self.shift}


class AddOne(zntrack.Node):
    """Add one to the number."""

    number: int = zntrack.deps()
    outs: int = zntrack.outs()

    def run(self) -> None:
        """Add one to the number."""
        self.outs = self.number + 1


class WriteDVCOuts(zntrack.Node):
    """Write an output file."""

    params: t.Any = zntrack.params()
    outs: pathlib.Path | str = zntrack.outs_path(zntrack.nwd / "output.txt")

    def run(self):
        """Write an output file."""
        pathlib.Path(self.outs).write_text(str(self.params))

    def get_outs_content(self):
        """Get the output file."""
        with self.state.use_tmp_path():
            return pathlib.Path(self.outs).read_text()


class WriteDVCOutsSequence(zntrack.Node):
    """Write an output file."""

    params: list = zntrack.params()
    outs: t.Union[list, tuple, set, dict] = zntrack.outs_path()

    def run(self):
        """Write an output file."""
        for value, path in zip(self.params, self.outs):
            pathlib.Path(path).write_text(str(value))

    def get_outs_content(self):
        """Get the output file."""
        data = []
        with self.state.use_tmp_path():
            for path in self.outs:
                data.append(pathlib.Path(path).read_text())
        return data


class WriteDVCOutsPath(zntrack.Node):
    """Write an output file."""

    params: t.Any = zntrack.params()
    outs: pathlib.Path | str = zntrack.outs_path(zntrack.nwd / "data")

    def run(self):
        """Write an output file."""
        pathlib.Path(self.outs).mkdir(parents=True, exist_ok=True)
        (pathlib.Path(self.outs) / "file.txt").write_text(str(self.params))

    def get_outs_content(self):
        """Get the output file."""
        with self.state.use_tmp_path():
            try:
                return (pathlib.Path(self.outs) / "file.txt").read_text()
            except FileNotFoundError:
                files = list(pathlib.Path(self.outs).iterdir())
                raise ValueError(f"Expected {self.outs} file, found {files}.")


class WriteMultipleDVCOuts(zntrack.Node):
    """Write an output file."""

    params: t.Any = zntrack.params()
    outs1: pathlib.Path = zntrack.outs_path(zntrack.nwd / "output.txt")
    outs2: pathlib.Path = zntrack.outs_path(zntrack.nwd / "output2.txt")
    outs3: pathlib.Path = zntrack.outs_path(zntrack.nwd / "data")

    def run(self):
        """Write an output file."""
        pathlib.Path(self.outs1).write_text(str(self.params[0]))
        pathlib.Path(self.outs2).write_text(str(self.params[1]))
        pathlib.Path(self.outs3).mkdir(parents=True, exist_ok=True)
        (pathlib.Path(self.outs3) / "file.txt").write_text(str(self.params[2]))

    def get_outs_content(self) -> t.Tuple[str, str, str]:
        """Get the output file."""
        with self.state.use_tmp_path():
            outs1_content = pathlib.Path(self.outs1).read_text()
            outs2_content = pathlib.Path(self.outs2).read_text()
            outs3_content = (pathlib.Path(self.outs3) / "file.txt").read_text()
            return outs1_content, outs2_content, outs3_content


class ComputeRandomNumber(zntrack.Node):
    """Compute a random number."""

    params_file: str = zntrack.params_path()

    number: float = zntrack.outs()

    # def _post_init_(self):
    #     super()._post_init_()
    #     self.params_file = pathlib.Path(self.params_file)
    #     raise ValueError("This is a test exception, simulating killing the Node.")

    def run(self):
        """Compute a random number."""
        self.number = self.get_random_number()

    def get_random_number(self):
        """Compute a random number."""
        params = json.loads(pathlib.Path(self.params_file).read_text())
        random.seed(params["seed"])
        return random.randint(params["min"], params["max"])

    def write_params(self, min, max, seed):
        """Write params to file."""
        pathlib.Path(self.params_file).write_text(
            json.dumps({"min": min, "max": max, "seed": seed})
        )


class ComputeRandomNumberWithParams(zntrack.Node):
    """Compute a random number."""

    min: int = zntrack.params()
    max: int = zntrack.params()
    seed: int = zntrack.params()

    number: float = zntrack.outs()

    def run(self):
        """Compute a random number."""
        self.number = self.get_random_number()

    def get_random_number(self):
        """Compute a random number."""
        random.seed(self.seed)
        return random.randint(self.min, self.max)


class ComputeRandomNumberNamed(ComputeRandomNumber):
    """Same as ComputeRandomNumber but with a custom name."""

    _name_ = "custom_ComputeRandomNumber"


class SumRandomNumbers(zntrack.Node):
    """Sum a list of random numbers."""

    numbers: t.List[ComputeRandomNumber] = zntrack.deps()
    result: int = zntrack.outs()

    def run(self):
        """Sum a list of random numbers."""
        self.result = sum(x.get_random_number() for x in self.numbers)


class SumRandomNumbersNamed(SumRandomNumbers):
    """Same as SumRandomNumbers but with a custom name."""

    _name_ = "custom_SumRandomNumbers"


class NodeWithRestart(zntrack.Node):
    """Node that restarts."""

    start: int = zntrack.params()
    raise_exception_until: int = zntrack.params(0)

    count: int = zntrack.outs()

    def run(self) -> None:
        """Run the node.

        Increments the count by one, for each run.
        Check that the restart flag is set.
        """
        self.count = self.start + self.state.run_count
        if self.state.run_count > 1:
            assert self.state.restarted
        if self.state.run_count < self.raise_exception_until:
            raise ValueError("This is a test exception, simulating killing the Node.")


class OptionalDeps(zntrack.Node):
    """Node with optional dependencies."""

    value: float = zntrack.deps(None)
    result: float = zntrack.outs()

    def run(self) -> None:
        """Run the node."""
        self.result = self.value or 0.0



import pytest

import zntrack


class ExternalToOutput(zntrack.Node):
    node: zntrack.Node = zntrack.deps()
    attr: str = zntrack.params()
    attr_is_func: bool = zntrack.params(False)
    outs: str = zntrack.outs()

    def run(self):
        if self.attr_is_func:
            self.outs = getattr(self.node, self.attr)()
        else:
            self.outs = getattr(self.node, self.attr)


@pytest.mark.xfail(reason="pending implementation")
@pytest.mark.needs_internet
def test_external(proj_path):
    node = zntrack.from_rev(
        "HelloWorld",
        remote="https://github.com/PythonFZ/ZnTrackExamples.git",
        rev="890c714",
    )

    with zntrack.Project() as proj:
        node = ExternalToOutput(node=node, attr="random_number")

    proj.run()
    # node.load()
    assert node.outs == 123


@pytest.mark.xfail(reason="pending implementation")
@pytest.mark.needs_internet
def test_external_grp(proj_path):
    node = zntrack.from_rev(
        "HelloWorld",
        remote="https://github.com/PythonFZ/ZnTrackExamples.git",
        rev="890c714",
    )

    proj = zntrack.Project()
    with proj.group("test"):
        node = ExternalToOutput(node=node, attr="random_number")

    proj.run()
    # node.load()
    assert node.outs == 123



import zntrack.group


class MyNode(zntrack.Node):
    pass


def test_project_group(proj_path):
    proj = zntrack.Project()
    with proj.group() as grp:
        n = MyNode()

    assert isinstance(grp, zntrack.group.Group)
    assert grp.names == ("Group1",)
    assert n in grp
    assert n.state.group == grp
    assert len(grp) == 1
    assert n.name == "Group1_MyNode"



import pathlib

import h5py
import numpy as np
import numpy.testing as npt

import zntrack
from zntrack.config import NOT_AVAILABLE, FieldTypes

CWD = pathlib.Path(__file__).parent.resolve()


def _h5data_getter(self: zntrack.Node, name: str, suffix: str):
    file = (self.nwd / name).with_suffix(suffix)
    with self.state.fs.open(file, mode="rb") as f:
        return h5py.File(f, "r")["data"][()]


# TODO: if ZNTRACK_FIELD_SUFFIX is None (not, not defined!) then
#   assume a directory is being saved there for DVC


def _h5data_save_func(self: zntrack.Node, name: str, suffix: str):
    file = (self.nwd / name).with_suffix(suffix)
    with h5py.File(file, "w") as f:
        f.create_dataset("data", data=getattr(self, name))


def h5data(*, cache: bool = True, independent: bool = False, **kwargs):
    return zntrack.field(
        cache=cache,
        independent=independent,
        field_type=FieldTypes.OUTS,
        load_fn=_h5data_getter,
        dump_fn=_h5data_save_func,
        suffix=".h5",
        repr=False,
        init=False,
        default=NOT_AVAILABLE,
        **kwargs,
    )


class NumpyNode(zntrack.Node):
    content: np.ndarray = h5data()

    def run(self):
        self.content = np.arange(9).reshape(3, 3)


def test_np_node(proj_path):
    with zntrack.Project() as project:
        node = NumpyNode()
    project.repro()

    npt.assert_array_equal(node.content, np.arange(9).reshape(3, 3))



import contextlib
import dataclasses
import datetime
import json
import logging
import pathlib
import typing as t
import uuid
import warnings

import typing_extensions as ty_ex
import znfields
import znflow
from dvc.stage.exceptions import InvalidStageName
from dvc.stage.utils import is_valid_name

from zntrack.group import Group
from zntrack.state import NodeStatus
from zntrack.utils.misc import get_plugins_from_env

from .config import NOT_AVAILABLE, ZNTRACK_LAZY_VALUE, NodeStatusEnum

try:
    from typing import dataclass_transform
except ImportError:
    from typing_extensions import dataclass_transform

T = t.TypeVar("T", bound="Node")

log = logging.getLogger(__name__)


def _name_setter(self, attr_name: str, value: str) -> None:
    """Check if the node name is valid."""

    if value is not None and not is_valid_name(value):
        raise InvalidStageName

    if value is not None and "_" in value:
        warnings.warn(
            "Node name should not contain '_'."
            " This character is used for defining groups."
        )

    self.__dict__[attr_name] = value


def _name_getter(self, attr_name: str) -> str:
    """Retrieve the name of a node based on the current graph context.

    Parameter
    ---------
        attr_name (str): The attribute name to retrieve.

    Returns
    -------
        str: The resolved node name.

    """
    value = self.__dict__.get(attr_name)  # Safer lookup with .get()
    graph = znflow.get_graph()

    # If value exists and the graph is either empty or not inside a group, return it
    if value is not None:
        if graph is znflow.empty_graph or graph.active_group is None:
            return str(value)

    # If no graph is active, return the class name as the default
    if graph is znflow.empty_graph:
        return str(self.__class__.__name__)

    # Compute name based on project-wide node names
    return str(graph.compute_all_node_names()[self.uuid])


@dataclass_transform()
@dataclasses.dataclass(kw_only=True)
class Node(znflow.Node, znfields.Base):
    """A Node."""

    name: str | None = znfields.field(
        default=None, getter=_name_getter, setter=_name_setter
    )
    always_changed: bool = dataclasses.field(default=False, repr=False)

    _protected_ = znflow.Node._protected_ + ["nwd", "name", "state"]

    def __post_init__(self):
        if self.name is None:
            # automatic node names expects the name to be None when
            # exiting the graph context.
            if not znflow.get_graph() is not znflow.empty_graph:
                self.name = self.__class__.__name__

    def _post_load_(self):
        """Called after `from_rev` is called."""
        raise NotImplementedError

    def run(self):
        raise NotImplementedError

    def save(self):
        for plugin in self.state.plugins.values():
            with plugin:
                for field in dataclasses.fields(self):
                    value = getattr(self, field.name)
                    if any(value is x for x in [ZNTRACK_LAZY_VALUE, NOT_AVAILABLE]):
                        raise ValueError(
                            f"Field '{field.name}' is not set."
                            " Please set it before saving."
                        )
                    try:
                        plugin.save(field)
                    except Exception as err:  # noqa: E722
                        if plugin._continue_on_error_:
                            warnings.warn(
                                f"Plugin {plugin.__class__.__name__} failed to"
                                f" save field {field.name}."
                            )
                        else:
                            raise err

        _ = self.state
        self.__dict__["state"]["state"] = NodeStatusEnum.FINISHED

    def __init_subclass__(cls):
        return dataclasses.dataclass(cls, kw_only=True)

    @property
    def nwd(self) -> pathlib.Path:
        return self.state.nwd

    @classmethod
    def from_rev(
        cls: t.Type[T],
        name: str | None = None,
        remote: str | None = None,
        rev: str | None = None,
        running: bool = False,
        lazy_evaluation: bool = True,
        **kwargs,
    ) -> T:
        if name is None:
            name = cls.__name__
        lazy_values = {}
        for field in dataclasses.fields(cls):
            # check if the field is in the init
            if field.init:
                lazy_values[field.name] = ZNTRACK_LAZY_VALUE

        lazy_values["name"] = name
        lazy_values["always_changed"] = None  # TODO: read the state from dvc.yaml
        instance = cls(**lazy_values)

        # TODO: check if the node is finished or not.
        instance.__dict__["state"] = NodeStatus(
            remote=remote,
            rev=rev,
            state=NodeStatusEnum.RUNNING if running else NodeStatusEnum.FINISHED,
            lazy_evaluation=lazy_evaluation,
            group=Group.from_nwd(instance.nwd),
        ).to_dict()

        instance.__dict__["state"]["plugins"] = get_plugins_from_env(instance)

        with contextlib.suppress(FileNotFoundError):
            # need to update run_count after the state is set
            # TODO: do we want to set the UUID as well?
            # TODO: test that run_count is correct, when using from_rev from another
            #  commit
            with instance.state.fs.open(instance.nwd / "node-meta.json") as f:
                content = json.load(f)
                run_count = content.get("run_count", 0)
                run_time = content.get("run_time", 0)
                if node_uuid := content.get("uuid", None):
                    instance._uuid = uuid.UUID(node_uuid)
                instance.__dict__["state"]["run_count"] = run_count
                instance.__dict__["state"]["run_time"] = datetime.timedelta(
                    seconds=run_time
                )

        if not instance.state.lazy_evaluation:
            for field in dataclasses.fields(cls):
                _ = getattr(instance, field.name)

        instance._external_ = True
        if not running and hasattr(instance, "_post_load_"):
            with contextlib.suppress(NotImplementedError):
                instance._post_load_()

        return instance

    @property
    def state(self) -> NodeStatus:
        if "state" not in self.__dict__:
            self.__dict__["state"] = NodeStatus().to_dict()
            self.__dict__["state"]["plugins"] = get_plugins_from_env(self)

        return NodeStatus(**self.__dict__["state"], node=self)

    @ty_ex.deprecated("loading is handled automatically via lazy evaluation")
    def load(self):
        pass



import typing as t

import pytest

import zntrack


class NodeWithMeta(zntrack.Node):
    author = zntrack.meta.Text("Fabian")
    title = zntrack.meta.Text("Test Node")


class NodeWithEnv(zntrack.Node):
    OMP_NUM_THREADS = zntrack.meta.Environment("1")

    result: t.Any = zntrack.outs()

    def run(self):
        import os

        assert os.environ["OMP_NUM_THREADS"] == self.OMP_NUM_THREADS, (
            f"{os.environ['OMP_NUM_THREADS']} != {self.OMP_NUM_THREADS}"
        )

        self.result = os.environ["OMP_NUM_THREADS"]


class NodeWithEnvNone(zntrack.Node):
    OMP_NUM_THREADS = zntrack.meta.Environment(None)

    result = zntrack.outs()

    def run(self):
        import os

        assert "OMP_NUM_THREADS" not in os.environ


class AssertGlobalEnv(zntrack.Node):
    def run(self):
        import os

        assert os.environ["ZNTRACK_EXAMPLE"] == "1"


class NodeWithEnvParam(NodeWithEnv):
    OMP_NUM_THREADS = zntrack.meta.Environment("1", is_parameter=True)


def test_GlobalEnv(proj_path):
    with zntrack.Project() as proj:
        _ = AssertGlobalEnv()
    proj.run(environment={"ZNTRACK_EXAMPLE": "1"})


@pytest.mark.parametrize("eager", [True, False])
def test_NodeWithMeta(proj_path, eager):
    with zntrack.Project() as project:
        node_w_meta = NodeWithMeta()

    project.run(eager=eager)
    # if not eager:
    #     node_w_meta.load()

    assert node_w_meta.author == "Fabian"


def test_NodeWithEnvNone(proj_path):
    with zntrack.Project() as proj:
        _ = NodeWithEnvNone()  # the actual test is inside the run method.
    proj.run()


class CombinedNodeWithMeta(zntrack.Node):
    input: str = zntrack.params("Hello ")
    output: str = zntrack.outs()
    author: str = zntrack.meta.Text()

    def run(self):
        self.output = self.input + self.author


def test_CombinedNodeWithMeta(proj_path):
    with pytest.raises(TypeError):
        # should raise an error because author is missing as kwarg
        _ = CombinedNodeWithMeta()

    with zntrack.Project() as proj:
        node = CombinedNodeWithMeta(author="World")
    proj.run()
    assert CombinedNodeWithMeta.from_rev().output == "Hello World"
    assert CombinedNodeWithMeta.from_rev().author == "World"

    node.author = "there"
    proj.run()

    # changing the 'meta.Text' should not trigger running the model again
    assert CombinedNodeWithMeta.from_rev().output == "Hello World"
    assert CombinedNodeWithMeta.from_rev().author == "there"

    zntrack.utils.run_dvc_cmd(["repro", "-f"])
    # Forcing rerun should use the updated meta keyword.
    assert CombinedNodeWithMeta.from_rev().output == "Hello there"


@pytest.mark.parametrize("cls", [NodeWithEnv, NodeWithEnvParam])
def test_NodeWithEnv(proj_path, cls):
    with zntrack.Project(force=True, automatic_node_names=False) as proj:
        node = cls()  # the actual test is inside the run method.
    proj.run()

    # check that the env variable is set correctly
    # node.load()
    assert node.result == "1"

    with proj:
        node = cls(OMP_NUM_THREADS="2")  # the actual test is inside the run method.
    proj.run()

    # node.load()
    if cls == NodeWithEnvParam:
        # Parameter will cause rerun and the result is changed
        assert node.result == "2"
    else:
        # env is not a parameter and will not cause rerun
        assert node.result == "1"


class EnvAsDict(zntrack.Node):
    env: dict = zntrack.meta.Environment({"OMP_NUM_THREADS": "1"})
    SPECIAL_ENV = zntrack.meta.Environment()

    def run(self):
        import os

        assert self.env["OMP_NUM_THREADS"] == "1"
        assert os.environ["OMP_NUM_THREADS"] == self.env["OMP_NUM_THREADS"]
        assert os.environ["ZNTRACK_EXAMPLE"] == "1"
        assert os.environ["SPECIAL_ENV"] == "25"


def test_EnvAsDict(proj_path):
    with zntrack.Project() as proj:
        _ = EnvAsDict(SPECIAL_ENV="25")  # the actual test is inside the run method.
    proj.run(environment={"ZNTRACK_EXAMPLE": "1"})



import contextlib
import importlib
import importlib.util
import pathlib
import sys

import dvc.api
import dvc.repo
import dvc.stage


def from_rev(name: str, remote: str | None = None, rev: str | None = None):
    fs = dvc.api.DVCFileSystem(url=remote, rev=rev)
    with fs.repo as repo:
        # with dvc.repo.Repo(remote=remote, rev=rev) as repo:
        for stage in repo.index.stages:
            with contextlib.suppress(AttributeError):
                # only PipelineStages have a name attribute
                if stage.name == name:
                    cmd = stage.cmd
                    break
        else:
            raise ValueError(f"Stage {name} not found in {repo}")

    # cmd will be "zntrack run module.name --name ..."
    # and we need the module.name and --name part
    run_str = cmd.split()[2]
    name = cmd.split()[4]

    package_and_module, cls_name = run_str.rsplit(".", 1)

    sys.path.append(pathlib.Path.cwd().as_posix())
    if remote is not None:
        # check if remote is a path that exists
        if pathlib.Path(remote).exists():
            sys.path.append(remote)

    try:
        module = importlib.import_module(package_and_module)
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            f"No module found for '{package_and_module}'. The package might be available "
            f"via 'pip install {package_and_module}' or from the remote"
            f"via 'pip install git+{remote}'."
        )

    cls = getattr(module, cls_name)
    return cls.from_rev(name, remote=remote, rev=rev)



import subprocess

import numpy as np

import zntrack


class ParamsToNumpy(zntrack.Node):
    stop: int = zntrack.params()

    result: np.ndarray = zntrack.outs()

    def run(self):
        self.result = np.arange(self.stop)


def test_ParamsToNumpy(proj_path):
    with zntrack.Project() as proj:
        node = ParamsToNumpy(stop=10)

    proj.build()
    subprocess.check_call(["dvc", "repro"])

    node = zntrack.from_rev(node.name)
    assert np.all(node.result == np.arange(node.stop))



import typing as t

import pytest

import zntrack


class CheckType(zntrack.Node):
    params: t.Any = zntrack.params()

    def run(self):
        pass


@pytest.fixture
def arg(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def fix_list() -> list:
    return list(range(10))


@pytest.fixture
def fix_dict() -> dict:
    # keys must be of type str!
    return {str(key): val for key, val in enumerate(range(10))}


@pytest.fixture
def fix_int() -> int:
    return 42


@pytest.fixture
def fix_empty_list() -> list:
    return []


@pytest.mark.parametrize(
    "arg", ["fix_list", "fix_int", "fix_dict", "fix_empty_list"], indirect=True
)
def test_params(arg, proj_path):
    project = zntrack.Project()
    with project:
        node = CheckType(params=arg)

    project.build()

    assert node.params == arg
    assert CheckType.from_rev().params == arg



import datetime
import unittest.mock as mock

from znflow.deployment import VanillaDeployment


class ZnTrackDeployment(VanillaDeployment):
    def _run_node(self, node_uuid):
        node = self.graph.nodes[node_uuid]["value"]
        start_time = datetime.datetime.now()
        node.state.increment_run_count()
        node.state.save_node_meta()
        if hasattr(node, "_method"):
            method_string = getattr(node, "_method")
            method = getattr(node, method_string)
            if method != "run":
                # mock the node.run with the method
                with mock.patch.object(node, "run", method):
                    # TODO: this needs to be fixed on the znflow side!
                    super()._run_node(node_uuid)
            else:
                super()._run_node(node_uuid)
        else:
            super()._run_node(node_uuid)

        run_time = datetime.datetime.now() - start_time
        node.state.add_run_time(run_time)
        node.state.save_node_meta()
        node.save()

    # TODO: when finished all Nodes, commit all changes



from .base import ZnTrackPlugin, base_getter, get_exp_info, plugin_getter, set_exp_info

__all__ = [
    "ZnTrackPlugin",
    "base_getter",
    "get_exp_info",
    "set_exp_info",
    "plugin_getter",
]



import enum
import pathlib

import typing_extensions as tyex

DVC_FILE_PATH = pathlib.Path("dvc.yaml")
PARAMS_FILE_PATH = pathlib.Path("params.yaml")
ZNTRACK_FILE_PATH = pathlib.Path("zntrack.json")
ENV_FILE_PATH = pathlib.Path("env.yaml")
NWD_PATH = pathlib.Path("nodes")
EXP_INFO_PATH = pathlib.Path(".exp_info.yaml")


# For "node-meta.json" and "dvc stage add ... --metrics-no-cache" the default is using
# git tracked files. Setting this to True will override the default behavior to always
# use the DVC cache. If you have a DVC cache setup, this might be desirable, to avoid
# a mixture between DVC cache and git tracked files.
ALWAYS_CACHE: bool = False


# Use sentinel object for zntrack specific configurations. Use
# a class to give it a better repr.
class _FIELD_TYPE_TYPE:
    pass


FIELD_TYPE = _FIELD_TYPE_TYPE()


class _ZNTRACK_CACHE_TYPE:
    pass


ZNTRACK_CACHE = _ZNTRACK_CACHE_TYPE()


class _ZNTRACK_SAVE_FUNC_TYPE:
    pass


ZNTRACK_SAVE_FUNC = _ZNTRACK_SAVE_FUNC_TYPE()


class _NOT_AVAILABLE_TYPE:
    def __repr__(self) -> str:
        return "NOT_AVAILABLE"


NOT_AVAILABLE = _NOT_AVAILABLE_TYPE()


class _ZNTRACK_LAZY_VALUE_TYPE:
    pass


ZNTRACK_LAZY_VALUE = _ZNTRACK_LAZY_VALUE_TYPE()


class _ZNTRACK_EMPTY_RETRUN_VALUE_TYPE:
    pass


PLUGIN_EMPTY_RETRUN_VALUE = _ZNTRACK_EMPTY_RETRUN_VALUE_TYPE()


class _ZNTRACK_INDEPENDENT_OUTPUT_TYPE:
    pass


ZNTRACK_INDEPENDENT_OUTPUT_TYPE = _ZNTRACK_INDEPENDENT_OUTPUT_TYPE()


class _ZNTRACK_OPTION_PLOTS_CONFIG:
    pass


ZNTRACK_OPTION_PLOTS_CONFIG = _ZNTRACK_OPTION_PLOTS_CONFIG()


class _ZNTRACK_FIELD_LOAD_TYPE:
    pass


# TODO: rename getter to align with dump to load
ZNTRACK_FIELD_LOAD = _ZNTRACK_FIELD_LOAD_TYPE()


class _ZNTRACK_FIELD_DUMP_TYPE:
    pass


ZNTRACK_FIELD_DUMP = _ZNTRACK_FIELD_DUMP_TYPE()


class _ZNTRACK_FIELD_SUFFIX_TYPE:
    pass


ZNTRACK_FIELD_SUFFIX = _ZNTRACK_FIELD_SUFFIX_TYPE()


class NodeStatusEnum(enum.Enum):
    CREATED = 0
    RUNNING = 2
    FINISHED = 3


class FieldTypes(str, enum.Enum):
    DEPS = "deps"
    PARAMS = "params"
    OUTS = "outs"
    PLOTS = "plots"
    METRICS = "metrics"

    PARAMS_PATH = "params_path"
    DEPS_PATH = "deps_path"
    OUTS_PATH = "outs_path"
    PLOTS_PATH = "plots_path"
    METRICS_PATH = "metrics_path"


@tyex.deprecated("ZnTrackOptionEnum is deprecated. Use FieldTypes instead.")
class ZnTrackOptionEnum(str, enum.Enum):
    DEPS = "deps"
    PARAMS = "params"
    OUTS = "outs"
    PLOTS = "plots"
    METRICS = "metrics"

    PARAMS_PATH = "params_path"
    DEPS_PATH = "deps_path"
    OUTS_PATH = "outs_path"
    PLOTS_PATH = "plots_path"
    METRICS_PATH = "metrics_path"



import json
import pathlib

import yaml
import znflow

import zntrack.config

CWD = pathlib.Path(__file__).parent.resolve()


class NodeA(zntrack.Node):
    results: int = zntrack.outs()

    def run(self):
        pass


class NodeB(zntrack.Node):
    input: int = zntrack.deps()

    def run(self):
        pass


def test_deps(proj_path):
    with zntrack.Project() as project:
        a1 = NodeA()
        a2 = NodeA()
        b = NodeB(input=a1.results + a2.results)

    assert isinstance(b.input, znflow.CombinedConnections)
    assert b.input.connections[0].instance == a1
    assert b.input.connections[1].instance == a2

    assert a1.results is zntrack.config.NOT_AVAILABLE
    assert a2.results is zntrack.config.NOT_AVAILABLE

    project.build()

    assert json.loads(
        (CWD / "zntrack_config" / "combined_dependencies.json").read_text()
    ) == json.loads((proj_path / "zntrack.json").read_text())
    assert yaml.safe_load(
        (CWD / "dvc_config" / "combined_dependencies.yaml").read_text()
    ) == yaml.safe_load((proj_path / "dvc.yaml").read_text())
    assert (CWD / "params_config" / "combined_dependencies.yaml").read_text() == (
        proj_path / "params.yaml"
    ).read_text()


if __name__ == "__main__":
    test_deps(None)



import contextlib
import dataclasses
import datetime
import importlib.metadata
import json
import pathlib
import tempfile
import typing as t
import warnings

import dvc.api
import dvc.repo
import dvc.stage.serialize
from dvc.utils import dict_sha256
from fsspec.implementations.local import LocalFileSystem
from fsspec.spec import AbstractFileSystem

from zntrack.config import NodeStatusEnum
from zntrack.group import Group
from zntrack.plugins import ZnTrackPlugin
from zntrack.utils.node_wd import get_nwd

if t.TYPE_CHECKING:
    from zntrack import Node

PLUGIN_LIST = list[t.Type[ZnTrackPlugin]]
PLUGIN_DICT = dict[str, ZnTrackPlugin]


@dataclasses.dataclass(frozen=True)
class NodeStatus:
    """Node status object.

    Parameters
    ----------
    remote : str, optional
        The repository remote, e.g. the URL of the git repository.
    rev : str, optional
        The revision of the repository, e.g. the git commit hash.
    run_count : int
        How often this Node has been run.
        Only incremented when the Node is restarted.
    state : NodeStatusEnum
        The state of the Node.
    lazy_evaluation : bool
        Whether to load fields lazily.
    tmp_path : pathlib.Path, optional
        The temporary path when using 'use_tmp_path'.
    node : Node, optional
        The Node object.
    plugins : dict[str, ZnTrackPlugin], optional
        Active plugins. In addition to the default plugins, MLFLow or AIM plugins will
        be added here.
    group : Group, optional
        The group of the Node.
    run_time : datetime.timedelta, optional
        The total run time of the Node.
    name: str
        The name of the Node.
    nwd: pathlib.Path
        The node working directory.
    restarted: bool
        Whether the Node was restarted and has been run at least once before.
    """

    remote: str | None = None
    rev: str | None = None
    run_count: int = 0
    state: NodeStatusEnum = NodeStatusEnum.CREATED
    lazy_evaluation: bool = True
    tmp_path: pathlib.Path | None = None
    node: "Node|None" = dataclasses.field(
        default=None, repr=False, compare=False, hash=False
    )
    plugins: PLUGIN_DICT = dataclasses.field(
        default_factory=dict, compare=False, repr=False
    )
    group: Group | None = None
    run_time: datetime.timedelta | None = None
    # TODO: move node name and nwd to here as well

    @property
    def name(self) -> str:
        return self.node.name

    @property
    def nwd(self):
        if self.tmp_path is not None:
            return self.tmp_path
        return get_nwd(self.node)

    @property
    def fs(self) -> AbstractFileSystem:
        """Get the file system of the Node.

        If the remote is None, the local file system is returned.
        Otherwise, a DVCFileSystem is returned.
        The FileSystem should be used to open files to ensure,
        that the correct version of the file is loaded.

        Examples
        --------

        >>> import zntrack
        >>> from pathlib import Path
        >>>
        >>> class MyNode(zntrack.Node):
        ...     outs_path: Path = zntrack.outs_path(zntrack.nwd / "file.txt")
        ...
        ...     def run(self):
        ...         self.outs_path.parent.mkdir(parents=True, exist_ok=True)
        ...         self.outs_path.write_text("Hello World!")
        ...
        ...     @property
        ...     def data(self):
        ...         with self.state.fs.open(self.outs_path) as f:
        ...             return f.read()
        ...
        >>> # build and run the graph and make multiple commits.
        >>> # the filesystem ensures that the correct version of the file is loaded.
        >>>
        >>> zntrack.from_rev("MyNode", rev="HEAD").data
        >>> zntrack.from_remote("MyNode", rev="HEAD~1").data

        """
        if self.remote is None and self.rev is None:
            return LocalFileSystem()
        return dvc.api.DVCFileSystem(
            url=self.remote,
            rev=self.rev,
        )

    @property
    def dvc_fs(self) -> dvc.api.DVCFileSystem:
        """Get the file system of the Node."""
        return dvc.api.DVCFileSystem(
            url=self.remote,
            rev=self.rev,
        )

    @property
    def restarted(self) -> bool:
        """Whether the node was restarted."""
        return self.run_count > 1

    @contextlib.contextmanager
    def use_tmp_path(self, path: pathlib.Path | None = None) -> t.Iterator[pathlib.Path]:
        """Load the data for ``*_path`` into a temporary directory.

        If you can not use ``node.state.fs.open`` you can use
        this as an alternative. This will load the data into
        a temporary directory and then delete it afterwards.
        The respective paths ``node.*_path`` will be replaced
        automatically inside the context manager.

        This is only set, if either ``remote`` or ``rev`` are set.
        Otherwise, the data will be loaded from the current directory.

        Examples
        --------

        >>> import zntrack
        >>> from pathlib import Path
        >>>
        >>> class MyNode(zntrack.Node):
        ...     outs_path: Path = zntrack.outs_path(zntrack.nwd / "file.txt")
        ...
        ...     def run(self):
        ...         self.outs_path.parent.mkdir(parents=True, exist_ok=True)
        ...         self.outs_path.write_text("Hello World!")
        ...
        ...     @property
        ...     def data(self):
        ...         with self.state.use_tmp_path():
        ...             with open(self.outs_path) as f:
        ...                 return f.read()
        ...
        >>> # build and run the graph and make multiple commits.
        >>> # the `use_tmp_path` ensures that the correct version
        >>> # of the file is loaded in the temporary directory and
        >>> # the `self.outs_path` is updated accordingly.
        >>>
        >>> zntrack.from_rev("MyNode", rev="HEAD").data
        >>> zntrack.from_remote("MyNode", rev="HEAD~1").data

        """
        if path is not None:
            raise NotImplementedError("Custom paths are not implemented yet.")

        # This feature is only required when the load
        #  is loaded, not when it is saved/executed
        if self.remote is None and self.rev is None:
            warnings.warn(
                "The temporary path is not used when neither remote or rev are set."
                "Consider checking for `self.state.remote` and `self.state.rev` when"
                "using `with node.state.use_tmp_path(): ...` ."
            )
            yield pathlib.Path.cwd()
            return

        with tempfile.TemporaryDirectory() as tmpdir:
            self.node.__dict__["state"]["tmp_path"] = pathlib.Path(tmpdir)
            try:
                yield pathlib.Path(tmpdir)
            finally:
                self.node.__dict__["state"].pop("tmp_path")

    def get_stage(self) -> dvc.stage.Stage:
        """Access to the internal dvc.repo api."""
        stage = next(iter(self.dvc_fs.repo.stage.collect(self.name)))
        if self.rev is None and self.remote is None:
            # If we want to look at the current workspace result, we need to
            # load all the information, not just dvc.yaml
            stage.save(allow_missing=True, run_cache=False)
        return stage

    def get_stage_lock(self) -> dict:
        """Access to the internal dvc.repo api."""
        stage = self.get_stage()
        return dvc.stage.serialize.to_single_stage_lockfile(stage)

    def get_stage_hash(self, include_outs: bool = False) -> str:
        """Get the hash of the stage."""
        stage_lock = self.get_stage_lock()

        if include_outs:
            return dict_sha256(stage_lock)
        else:
            filtered_lock = {
                k: v for k, v in stage_lock.items() if k in ["cmd", "deps", "params"]
            }
            return dict_sha256(filtered_lock)

    def to_dict(self) -> dict:
        """Convert the NodeStatus to a dictionary."""
        content = dataclasses.asdict(self)
        content.pop("node")
        return content

    def get_field(self, attribute: str) -> dataclasses.Field:
        fields = dataclasses.fields(self.node)
        for field in fields:
            if field.name == attribute:
                return field
        else:
            raise AttributeError(f"Unable to locate '{attribute}' in {self.node}.")

    def add_run_time(self, run_time: datetime.timedelta) -> None:
        """Add the run time to the node."""
        if self.run_time is None:
            self.node.__dict__["state"]["run_time"] = run_time
        else:
            self.node.__dict__["state"]["run_time"] += run_time

    def increment_run_count(self) -> None:
        self.node.__dict__["state"]["run_count"] = self.run_count + 1

    def save_node_meta(self) -> None:
        node_meta_content = {
            "uuid": str(self.node.uuid),
            "run_count": self.run_count,
            "zntrack_version": importlib.metadata.version("zntrack"),
        }

        if self.run_time is not None:
            node_meta_content["run_time"] = self.run_time.total_seconds()

        with contextlib.suppress(importlib.metadata.PackageNotFoundError):
            module = self.node.__module__.split(".")[0]
            node_meta_content["package_version"] = importlib.metadata.version(module)
        self.nwd.mkdir(parents=True, exist_ok=True)
        (self.nwd / "node-meta.json").write_text(json.dumps(node_meta_content, indent=2))
